/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ACTIONTREEITEM_H
#define ACTIONTREEITEM_H

#include <qtreeitem.h>

#include <QtXml/QDomNode>

namespace NSCore {

	/**
	 * @brief Action in a tree-like structure
	 */
	class ActionTreeItem : public QTreeItem
	{
		Q_OBJECT
		/**
		 * @brief Path to the icon
		 *
		 * Can be either a real system-path or a relativ-qresource-path
		 */
		Q_PROPERTY( QString iconPath READ iconPath WRITE setIconPath )

		public:
			/**
			 * @brief Creates a new action
			 *
			 * @param n the name of this action
			 * @param p the parent QTreeItem
			 * @param i the path to the icon
			 */
			ActionTreeItem( QString n, QTreeItem* p, QString i="" );

			/**
			 * @brief Creates a new action
			 *
			 * @param n the name of this action
			 * @param p the parent QTreeItem
			 * @param o the object to send the signal to
			 * @param s the slot to connect to
			 */
			ActionTreeItem( QString n, QTreeItem* p, QObject* o, const char* s );

			/**
			 * @brief Creates a new action
			 *
			 * @param n the name of this action
			 * @param p the parent QTreeItem
			 * @param o the object to send the signal to
			 * @param s the slot to connect to
			 * @param i the path to the icon
			 */
			ActionTreeItem( QString n, QTreeItem* p, QString i, QObject* o, const char* s );

			/**
			 * @brief Copy a ActionTreeItem and possibly its childs
			 * into another tree
			 *
			 * Childs are only copied if they are also ActionTreeItems...
			 *
			 * @param item The item that is to be copied
			 * @param p The parent of the new item
			 */
			ActionTreeItem( ActionTreeItem& item, QTreeItem* p );

			/// destructor
			~ActionTreeItem();

			/// Returns the guiHint
			int guiHint() const { return _guihint; }
			/// Set the guiHint
			void setGuiHint( int n ) { _guihint = n; }

			/**
			 * @brief adds a new Parameter as child
			 *
			 * @param path The path to the new item.
			 * @param tooltip A string to be shown as tooltip in GUIs.
			 */
			void addAction( QString path, QString tooltip=0 );

			// Gui-Functions from QTreeItem:
			QVariant guiValue() const { return QVariant(); }
			void guiSetValue( QVariant ) { }
			Qt::ItemFlags guiFlags() const;
			QStringList guiActions() const;
			void guiExecAction( QString, GuiServer_Interface* );
			QVariant guiIdentifier( int role ) const;

			/**
			 * @brief Return the path of the icon for this action
			 */
			QString iconPath() const { return _iconpath; }
			/**
			 * @brief Set the path of the icon for this action
			 */
			void setIconPath( QString n ) { _iconpath = n; }

		public slots:
			/**
			 * @brief Execute the action from scripts/etc.
			 */
			void exec();

		signals:
			/**
			 * @brief Execute the desired action
			 *
			 * Connect this signal to your slot that should get executed by the gui.
			 */
			void activated();

		private:
			// Debug
			QString debug() const {
				return QString( "Action: %1" ).arg( objectName() );
			}

			QDomNode saveToXML( QDomDocument );
			bool needsToSaveToXml() { return false; }
			bool restoreFromXML( QDomNode );

			/**
			 * @brief masked copy-constructor
			 *
			 * has to be private since the new parent needs to be known
			 */
			ActionTreeItem( const ActionTreeItem& )
				: QTreeItem( "", 0 )
				{}

			/// Internal storage of the guihint
			int _guihint;
			/// Internal storage of the iconpath
			QString _iconpath;
	};

};

#endif // ACTIONTREEITEM_H

