/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "datasetchain.h"
#include "datasetchain.moc"

#include <QtCore/QDebug>
#include <QtCore/QThread>
#include <QtCore/QCoreApplication>
#include <QtCore/QTime>
#include <QtCore/QTimer>
#include <QtCore/QReadWriteLock>
#include <QtCore/QReadLocker>
#include <QtCore/QWriteLocker>
#include <QtCore/QEvent>

namespace NSCore {

	const QEvent::Type MyEvent = (  QEvent::Type )1003;
	class DataSetEvent : public QEvent
	{
		public:
			DataSetEvent( AcqisitionDataSet d ) : QEvent( MyEvent ), data( d ) {
				//qDebug() << "DataSetEvent::DataSetEvent(" << d << ")";
			}
			AcqisitionDataSet data;
	};

	class DataSetProducer::DataSetProducerPrivate {
		public:
			QReadWriteLock* consumerslock;
			QList< Pointer_template<DataSetChainMember> > consumers;
			DataSetProducerPrivate()
			{
				consumerslock = new QReadWriteLock(/* QReadWriteLock::Recursive*/ );
				time = QTime::currentTime();
				time.start();
				counter = 0;
				dispatching = DispatchToAll;
				dispatchiterator = 0; //consumers.begin();
			}
			~DataSetProducerPrivate() {
				delete consumerslock;
			}

			int counter;
			QTime time;

			int dispatching;
			//QList< Pointer_template<DataSetChainMember> >::Iterator dispatchiterator;
			int dispatchiterator;
	};

	DataSetProducer::DataSetProducer( QObject* p ) : QTreeItem( p ), _d( new DataSetProducerPrivate() ) {
		//qDebug() << "DataSetProducer::DataSetProducer() living in thread" << QThread::currentThread();
	}
	DataSetProducer::~DataSetProducer() {
		//qDebug() << "DataSetProducer::~DataSetProducer()";
		QWriteLocker( _d->consumerslock );
		while ( ! _d->consumers.isEmpty() )
			_d->consumers.removeFirst();
		delete _d;
	}

	void DataSetProducer::registerConsumer( DataSetChainMember* consumer ) {
		//qDebug() << "DataSetProducer::registerConsumer(" << consumer << ")";
		if ( consumer == 0 )
			qWarning() << "Tried to register Consumer with address 0!";
		else {
			QWriteLocker( _d->consumerslock );
			_d->consumers.append( Pointer_template<DataSetChainMember>( consumer ) );
		}
	}
	void DataSetProducer::unregisterConsumer( DataSetChainMember* consumer ) {
		QWriteLocker( _d->consumerslock );
		foreach( Pointer_template<DataSetChainMember> pointer, _d->consumers ) {
			if ( pointer.object() == consumer )
				_d->consumers.removeAll( pointer );
		}
	}
	void DataSetProducer::unregisterAllConsumers() {
		while ( ! _d->consumers.isEmpty() )
			unregisterConsumer( _d->consumers.first().object() );
	}

	void DataSetProducer::setDispatching( Dispatching n ) {
		_d->dispatching = n;
	}

	void DataSetProducer::enqueue( AcqisitionDataSet& data ) {
		//qDebug() << "DataSetProducer::enqueue(" << &data << ")";
		if ( _d->time.elapsed() >= 1000 ) {
			emit countsPerSecond( _d->counter/( _d->time.elapsed()/1000.0 ) );
			_d->time.restart();
			_d->counter = 0;
		}
		if ( !( _d->consumers.isEmpty() ) ) {
			QReadLocker( _d->consumerslock );
			if ( _d->dispatching == DispatchEqual ) {
				if ( _d->dispatchiterator >= _d->consumers.size() )
					_d->dispatchiterator = 0;
				DataSetEvent* ev = new DataSetEvent( data );
				QCoreApplication::postEvent( _d->consumers.at( _d->dispatchiterator ).object(), ev );
				_d->dispatchiterator += 1;
			}
			if ( _d->dispatching == DispatchToAll ) {
				for ( int i=0; i<_d->consumers.size(); ++i ) {
					DataSetEvent* ev = new DataSetEvent( data );
					QCoreApplication::postEvent( _d->consumers.at( i ).object(), ev );
				}
			}
		}
		_d->counter++;
	}

	int DataSetProducer::consumersSize() const {
		QReadLocker( _d->consumerslock );
		return _d->consumers.size();
	}
	bool DataSetProducer::consumersIsEmpty() const {
		QReadLocker( _d->consumerslock );
		return _d->consumers.isEmpty();
	}


	DataSetChainMember::DataSetChainMember( QObject* p )
		: DataSetProducer( p )
	{
		//qDebug() << "DataSetChainMember::DataSetChainMember()";
	}
	DataSetChainMember::~DataSetChainMember() {
		//qDebug() << "DataSetChainMember::~DataSetChainMember()";
	}

	bool DataSetChainMember::event( QEvent* ev ) {
		if ( ev->type() == MyEvent ) {
			//qDebug() << "DataSetChainMember::event(" << ev << ") Got my own event-type!\n this=" << this;
			DataSetEvent* e = dynamic_cast<DataSetEvent*>( ev );
			if ( e ) {
				if ( dataArrived( e->data ) )
					enqueue( e->data );
			}
			return true;
		}
		return QObject::event( ev );
	}

};

