/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include <math.h>

#include "mathexpression.h"
#include "mathexpression_p.h"
#include "mathexpression.moc"

#include <QtCore/QTimer>

using namespace NSCore;

MathExpression::MathExpression( NSCore::QTreeItem* parent )
	: NSCore::QTreeItem( parent )
{
	setObjectName( "MathExpression" );
}

QString MathExpression::guiString() const { return objectName(); }
QStringList MathExpression::guiActions() const { return NSCore::QTreeItem::guiActions() << "Remove me"; }

void MathExpression::guiExecAction( QString name, NSCore::GuiServer_Interface* gui ) {
	//qDebug() << "MathExpression::guiExecAction(" << name << "," << gui << ")";
	if ( name == "Remove me" ) {
		deleteMe( this, gui );
		return;
	}
	if ( name == "Add Exp." ) {
		addChild( this, gui );
		return;
	}
	NSCore::QTreeItem::guiExecAction( name, gui );
}

void MathExpression::deleteMe( MathExpression* exp, NSCore::GuiServer_Interface* gui ) {
	//qDebug() << "MathExpression::deleteMe(" << exp << "," << gui << ")";
	MathExpression* p = dynamic_cast<MathExpression*>( parent() );
	if ( p )
		p->deleteMe( exp, gui );
}
void MathExpression::addChild( MathExpression* pexp, NSCore::GuiServer_Interface* gui ) {
	//qDebug() << "MathExpression::addChild(" << pexp << "," << gui << ")";
	MathExpression* p = dynamic_cast<MathExpression*>( parent() );
	if ( p )
		p->addChild( pexp, gui );
}

MathTreeRoot::MathTreeRoot( QTreeItem* p ) : MathExpression( p ), _singleChild( true ) {
	qDebug() << "MathTreeRoot::MathTreeRoot()";
	setObjectName( "Math" );
	QTimer* tmp = new QTimer( this );
	tmp->setInterval( 500 );
	connect( tmp, SIGNAL( timeout() ), this, SLOT( workOnCache() ) );
	tmp->start();
}
MathTreeRoot::~MathTreeRoot() {
	qDebug() << "MathTreeRoot::~MathTreeRoot()";
}
void MathTreeRoot::deleteMe( MathExpression* exp, GuiServer_Interface* gui ) {
	if ( exp == this )
		return;
	if ( gui->messageYesNo( "Really?", "Do you really want to delete this item?" ) )
		exp->deleteLater();
}
void MathTreeRoot::addChild( MathExpression* pexp, GuiServer_Interface* gui ) {
	qDebug() << "MathTreeRoot::addChild()";
	//qDebug() << " my thread     :" << thread();
	//qDebug() << " gui thread    :" << qApp->thread();
	//qDebug() << " current thread:" << QThread::currentThread();
	QList<GuiServer_Interface::ListItem> items;
	foreach( QString elem, MathExpressionFromXMLCreator::the()->availableElements() ) {
		GuiServer_Interface::ListItem item;
		item.title = elem;
		item.help = MathExpressionFromXMLCreator::the()->helpText( elem );
		items.append( item );
	}
	QString type = gui->getItem( "New MathExpression", "Select the new MathExpression to add:", items );
	if ( !type.isEmpty() ) {
		if ( pexp == this && _singleChild )
			foreach( MathExpression* item, childobjects<MathExpression>() )
				item->deleteLater();
		_childsToAdd.push_back( QPair<QString, MathExpression*>( type, pexp ) );
	}
}
void MathTreeRoot::workOnCache() {
	while ( ! _childsToAdd.isEmpty() ) {
		QPair<QString, MathExpression*> p = _childsToAdd.takeFirst();
		qDebug() << "MathTreeRoot::workOnCache() creating" << p.first;
		MathExpressionFromXMLCreator::the()->fromXML( QString( "<%1 />" ).arg( p.first ), p.second );
	}
}
double MathTreeRoot::value( AcqisitionDataSet& data ) {
	QList<MathExpression*> tmp = childobjects<MathExpression>();
	if ( ! tmp.isEmpty() )
		return tmp.first()->value( data );
	return NAN;
}
void MathTreeRoot::guiExecAction( QString action, GuiServer_Interface* gui ) {
	if ( action == "Set Expression" )
		addChild( this, gui );
}
QDomNode MathTreeRoot::saveToXML( QDomDocument doc ) {
	QDomElement elem = doc.createElement( "path" );
	elem.setAttribute( "name", objectName() );
	return elem;
}
bool MathTreeRoot::restoreFromXML( QDomNode node ) {
	return ( node.nodeName() == "path" && node.toElement().attribute( "name" ) == objectName() );
}
bool MathTreeRoot::createFromXML( QDomNode node ) {
	//qDebug() << "MathTreeRoot::createFromXML()";
	//debugPrint();
	while ( ! childobjects<NSCore::MathExpression>().isEmpty() ) {
		delete childobjects<NSCore::MathExpression>().first();
	}
	//debugPrint();
	QDomElement elem = node.firstChildElement();
	while ( !elem.isNull() ) {
		MathExpressionFromXMLCreator::the()->fromXML( elem, this );
		elem = elem.nextSiblingElement();
	}
	//debugPrint();
	return false;
}


MathExpressionFromXMLCreator::MathExpressionFromXMLCreator() {
	//qDebug() << "MathExpressionFromXMLCreator::MathExpressionFromXMLCreator()";
	the()->_creators.push_back( this );
}
MathExpressionFromXMLCreator* MathExpressionFromXMLCreator::the() {
	static MathExpressionFromXMLCreator* me = 0;
	if ( !me ) {
		//qDebug() << "MathExpressionFromXMLCreator::the() me is null, will create myself...";
		me = new MathExpressionFromXMLCreator( true );
		me->_creators.push_back( me );
	}
	return me;
}
MathExpression* MathExpressionFromXMLCreator::fromXML( QString xml, QTreeItem* pexp ) {
	QDomDocument doc( "bla" );
	doc.setContent( xml );
	return fromXML( doc, pexp );
}
MathExpression* MathExpressionFromXMLCreator::fromXML( QDomNode node, QTreeItem* pexp ) {
	//qDebug() << "MathExpressionFromXMLCreator::fromXML(" << &node << "," << pexp << ")";
	if ( node.nodeName() == "#document" )
		node = node.toDocument().documentElement();

	MathExpression* ret = 0;
	QList<MathExpressionFromXMLCreator*>::iterator it = _creators.begin();
	while ( ret == 0 && it != _creators.end() ) {
		//qDebug() << " Trying" << *it << "for next element.";
		ret = ( *it )->singleElementFromXML( node, pexp );
		it++;
	}
	QDomNode child = node.firstChild();
	while ( !child.isNull() ) {
		fromXML( child, ret );
		child = child.nextSibling();
	}
	return ret;
}
QStringList MathExpressionFromXMLCreator::availableElements() const {
	QStringList ret;
	foreach( MathExpressionFromXMLCreator* c, _creators )
		ret << c->elementsICanCreate();
	return ret;
}
QString MathExpressionFromXMLCreator::helpText( QString n ) const {
	QString ret;
	QList<MathExpressionFromXMLCreator*>::const_iterator it = _creators.begin();
	while ( ret.isNull() && it != _creators.end() ) {
		ret = ( *it )->myHelpText( n );
		it++;
	}
	return ret;
}

MathExpression* MathExpressionFromXMLCreator::singleElementFromXML( QDomNode node, QTreeItem* pexp ) {
	//qDebug() << "MathExpressionFromXMLCreator::singleElementFromXML(" << &node << "," << pexp << ")";

	MathExpression* returnvalue = 0;
	if ( node.nodeName() == "result" )
		returnvalue = new NSCore::MathExpression_Result( pexp );
	if ( node.nodeName() == "constant" )
		returnvalue = new NSCore::MathExpression_Constant( pexp );
	if ( node.nodeName() == "smb" )
		returnvalue = new NSCore::MathExpression_SmB( pexp );
	if ( node.nodeName() == "add" )
		returnvalue = new NSCore::MathExpression_Add( pexp );
	if ( node.nodeName() == "mul" )
		returnvalue = new NSCore::MathExpression_Mul( pexp );
	if ( node.nodeName() == "min" )
		returnvalue = new NSCore::MathExpression_Min( pexp );
	if ( node.nodeName() == "max" )
		returnvalue = new NSCore::MathExpression_Max( pexp );
	if ( node.nodeName() == "neg" )
		returnvalue = new NSCore::MathExpression_Neg( pexp );
	if ( node.nodeName() == "precision" )
		returnvalue = new NSCore::MathExpression_Precision( pexp );
	if ( node.nodeName() == "gtz" )
		returnvalue = new NSCore::MathExpression_GtZ( pexp );
	if ( node.nodeName() == "ltz" )
		returnvalue = new NSCore::MathExpression_LtZ( pexp );
	if ( node.nodeName() == "window" )
		returnvalue = new MathExpression_Window( pexp );
	if ( node.nodeName() == "fuzzywindow" )
		returnvalue = new MathExpression_FuzzyWindow( pexp );
	if ( returnvalue )
		returnvalue->fromXml( node );
	return returnvalue;
}
QStringList MathExpressionFromXMLCreator::elementsICanCreate() const {
	return QStringList() << "result" << "sampleparameter" << "constant" << "smb" << "precision" << "add" << "mul" << "neg" << "min" << "max" << "gtz" << "ltz" << "window" << "fuzzywindow";
};
QString MathExpressionFromXMLCreator::myHelpText( QString n ) const {
	qDebug() << "MathExpressionFromXMLCreator::myHelpText(" << n.toStdString().c_str() << ")";
	if ( n == "result" ) return "Get a results value from the data.";
	if ( n == "sampleparameter" ) return "Get a parameter from the data.";
	if ( n == "constant" ) return "A constant value.";
	if ( n == "smb" ) return "<b>Smaller-Minus-Bigger</b>: decide which channels value to substract by another result.";
	if ( n == "precision" ) return "Reduce the precision of a value.";
	if ( n == "add" ) return "Addition of values.";
	if ( n == "mul" ) return "Multiply values.";
	if ( n == "neg" ) return "Negate the value.";
	if ( n == "min" ) return "Return the minimum of a number of values.";
	if ( n == "max" ) return "Return the maximum of a number of values.";
	if ( n == "gtz" ) return "Pass the value if its greater than zero, otherwise pass NaN.";
	if ( n == "ltz" ) return "Pass the value if its less than zero, otherwise pass NaN.";
	if ( n == "window" ) return "Pass the value (or the second value if present) if its inside the given window, otherwise pass NaN (or the third value if present).";
	return "Help for this element is not yet implemented!";
	return "";
}


MathExpression_Result::MathExpression_Result( NSCore::QTreeItem* parent )
	: MathExpression( parent )
{
	//maximumChilds( 3 );
	setObjectName( "Result" );
}

double MathExpression_Result::value( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "MathExpression_Result::value( data )";
	double ret=NAN;
	QVariant* tmp = data.result( _name );
	//qDebug() << i << "\t tmp =" << tmp;
	if ( tmp )
		ret = tmp->toDouble();
	//qDebug() << " returning" << ret;
	return ret;
}


MathExpression_Constant::MathExpression_Constant( NSCore::QTreeItem* parent )
	: MathExpression( parent )
	, _value( 0.0 )
{
	setObjectName( "Constant" );
}

double MathExpression_Constant::value( NSCore::AcqisitionDataSet& data ) {
	return _value;
}


MathExpression_Add::MathExpression_Add( NSCore::QTreeItem* parent ) : MathExpression( parent ) {
	setObjectName( "Add" );
}

double MathExpression_Add::value( NSCore::AcqisitionDataSet& data ) {
	double ret = 0;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 )
			ret += exp->value( data );
	}
	return ret;
}


MathExpression_Mul::MathExpression_Mul( NSCore::QTreeItem* parent ) : MathExpression( parent ) {
	setObjectName( "Multiplication" );
}

double MathExpression_Mul::value( NSCore::AcqisitionDataSet& data ) {
	double ret = 1;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 )
			ret *= exp->value( data );
	}
	return ret;
}


MathExpression_Min::MathExpression_Min( NSCore::QTreeItem* parent ) : MathExpression( parent ) {
	setObjectName( "Minimum" );
}

double MathExpression_Min::value( NSCore::AcqisitionDataSet& data ) {
	double ret = INFINITY;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 ) {
			/*double val = exp->value( data );
			if ( val < ret )
				ret = val;*/
			ret = qMin( ret, exp->value( data ) );
		}
	}
	return ret;
}

MathExpression_Max::MathExpression_Max( NSCore::QTreeItem* parent ) : MathExpression( parent ) {
	setObjectName( "Maximum" );
}

double MathExpression_Max::value( NSCore::AcqisitionDataSet& data ) {
	double ret = -INFINITY;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 ) {
			/*double val = exp->value( data );
			if ( val < ret )
				ret = val;*/
			ret = qMax( ret, exp->value( data ) );
		}
	}
	return ret;
}



MathExpression_Neg::MathExpression_Neg( NSCore::QTreeItem* parent )
	: MathExpression( parent )
{
	//maximumChilds( 1 );
	setObjectName( "Negation" );
}

double MathExpression_Neg::value( NSCore::AcqisitionDataSet& data ) {
	double ret = NAN;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 )
			ret = - exp->value( data );
	}
	return ret;
}


MathExpression_Precision::MathExpression_Precision( NSCore::QTreeItem* p )
	: MathExpression( p )
	, _precision( 0.001 )
{
	//maximumChilds( 2 );
	setObjectName( "Precision" );
}

double MathExpression_Precision::value( NSCore::AcqisitionDataSet& data ) {
	double ret = NAN;
	foreach( QObject* item, childobjects() ) {
		MathExpression* exp = dynamic_cast<MathExpression*>( item );
		if ( exp != 0 )
			ret = exp->value( data );
	}
	ret = ret - fmod( ret, _precision );
	return ret;
}


MathExpression_SmB::MathExpression_SmB( NSCore::QTreeItem* p )
	: MathExpression( p )
	, _selection( new NSCore::NamedParameterTreeItem( "selection", "", this ) )
	, _calculus( new NSCore::NamedParameterTreeItem( "calculus", "", this ) )
{
	//maximumChilds( 3 );
	setObjectName( "SmB" );
}

double MathExpression_SmB::value( NSCore::AcqisitionDataSet& data ) {
qWarning() << "MathExpression_SmB::value() FIXME!";
#if 0
	double ret = NAN;
	double _sel1=NAN, _sel2=NAN, _cal1=NAN, _cal2=NAN;
	int gen_select = _generation_select->toInt();
	int gen_calc = _generation_calc->toInt();
	if ( gen_calc == -2000 )
		gen_calc = gen_select;
	QString sel = _selection->toString();
	QString cal = _calculus->toString();
	if ( data.channel( 0 ) != 0 ) {
		if ( data.channel( 0 )->result( sel, gen_select ) != 0 )
			_sel1 = data.channel( 0 )->result( sel, gen_calc )->toDouble();
		if ( data.channel( 0 )->result( cal, gen_select ) != 0 )
			_cal1 = data.channel( 0 )->result( cal, gen_calc )->toDouble();
	}
	if ( data.channel( 1 ) != 0 ) {
		if ( data.channel( 1 )->result( sel, gen_select ) != 0 )
			_sel2 = data.channel( 1 )->result( sel, gen_calc )->toDouble();
		if ( data.channel( 1 )->result( cal, gen_select ) != 0 )
			_cal2 = data.channel( 1 )->result( cal, gen_calc )->toDouble();
	}
	if ( _sel1 < _sel2 )
		ret = _cal1 - _cal2;
	else
		if ( _sel2 <= _sel1 )
			ret = _cal2 - _cal1;

	return ret;
#endif
	return NAN;
}


MathExpression_GtZ::MathExpression_GtZ( NSCore::QTreeItem* p )
	: MathExpression( p ) {}

double MathExpression_GtZ::value( NSCore::AcqisitionDataSet& data ) {
	double ret = NAN;
	if ( childs() > 1 ) {
		MathExpression* tmp = dynamic_cast<MathExpression*>( child( 0 ) );
		if ( tmp && tmp->value( data ) > 0 ) {
			tmp = dynamic_cast<MathExpression*>( child( 1 ) );
			if ( tmp )
				ret = tmp->value( data );
		} else {
			tmp = dynamic_cast<MathExpression*>( child( 2 ) );
			if ( tmp )
				ret = tmp->value( data );
		}
	}
	return ret;
}

MathExpression_LtZ::MathExpression_LtZ( NSCore::QTreeItem* p )
	: MathExpression( p ) {}

double MathExpression_LtZ::value( NSCore::AcqisitionDataSet& data ) {
	double ret = NAN;
	if ( childs() > 1 ) {
		MathExpression* tmp = dynamic_cast<MathExpression*>( child( 0 ) );
		if ( tmp && tmp->value( data ) < 0 ) {
			tmp = dynamic_cast<MathExpression*>( child( 1 ) );
			if ( tmp )
				ret = tmp->value( data );
		} else {
			tmp = dynamic_cast<MathExpression*>( child( 2 ) );
			if ( tmp )
				ret = tmp->value( data );
		}
	}
	return ret;
}

MathExpression_Window::MathExpression_Window( NSCore::QTreeItem* p )
	: MathExpression( p )
	, _min( new NamedParameterTreeItem( "minimum", -5.0, this ) )
	, _max( new NamedParameterTreeItem( "maximum", 5.0, this ) )
{
	setObjectName( "Window" );
}

double MathExpression_Window::value( NSCore::AcqisitionDataSet& data ) {
	double ret = NAN;
	double min = _min->toDouble();
	double max = _max->toDouble();
	QList<MathExpression*> childs = childobjects<MathExpression>();
	if ( childs.size() > 0 ) {
		double tmp = childs[ 0 ]->value( data );
		if ( childs.size() == 1 ) {
			if ( min < tmp && tmp < max )
				ret = tmp;
		} else {
			if ( min < tmp && tmp < max )
				ret = childs[ 1 ]->value( data );
			else
				if ( childs.size() > 2 )
					ret = childs[ 2 ]->value( data );
		}
	}
	return ret;
}


