/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MATHEXPRESSION_H
#define MATHEXPRESSION_H

#include <qtreeitem.h>
#include <datatypes.h>

namespace NSCore {

/**
 * @brief Baseclass for an expression
 *
 * Expressions are used to extract any kind of results from the
 * NSCore::AcqisitionDataSet respectively from the results saved in them.
 */
class MathExpression : public NSCore::QTreeItem
{
	Q_OBJECT
	public:
		/// simple tree-constructor
		MathExpression( NSCore::QTreeItem* parent );

		QString guiString() const;
		QStringList guiActions() const;

		void guiExecAction( QString name, NSCore::GuiServer_Interface* gui );

		/**
		 * @brief delete items in a thread-save way...
		 *
		 * ...by handing the action to the parent until it gets handled in a
		 * different way.
		 */
		virtual void deleteMe( MathExpression* exp, NSCore::GuiServer_Interface* gui );
		/**
		 * @brief add new items in a thread-save way...
		 *
		 * ...by handing the action to the parent until it gets handled in a
		 * different way.
		 */
		virtual void addChild( MathExpression* pexp, NSCore::GuiServer_Interface* gui );

		/**
		 * @brief Abstract function that does the work
		 *
		 * This has to be implemented to do the work.
		 */
		virtual double value( NSCore::AcqisitionDataSet& ) =0;

};


class MathTreeRoot : public MathExpression
{
	Q_OBJECT
	Q_PROPERTY( bool singleChild READ singleChild WRITE setSingleChild );
	public:
		MathTreeRoot( NSCore::QTreeItem* parent );
		~MathTreeRoot();

		void deleteMe( MathExpression* exp, NSCore::GuiServer_Interface* gui );

		void addChild( MathExpression* pexp, NSCore::GuiServer_Interface* gui );

		double value( AcqisitionDataSet& );

		QStringList guiActions() const { return QTreeItem::guiActions() << "Set Expression"; }

		void guiExecAction( QString, GuiServer_Interface* );

		void setSingleChild( bool n ) { _singleChild = n; }
		bool singleChild() const { return _singleChild; }

	private slots:
		void workOnCache();

	private:
		QString debug() const { return "MathTreeRoot"; }

		QDomNode saveToXML( QDomDocument doc );
		bool restoreFromXML( QDomNode node );
		bool createFromXML( QDomNode );

		QList< QPair<QString,NSCore::MathExpression*> > _childsToAdd;

		bool _singleChild;
};

/**
 * @brief Singleton to create MathExpression trees
 *
 * This MathExpressionFromXMLCreator is to provide one ( extendible ) way of
 * creating MathExpressions or maybe even QTreeItems. Basicly all the knowledge
 * to create MathExpressions is centralized here to provide a single api for all
 * users.
 *
 * It should be extendible in a way that derived creators register themself at
 * the global creator and are then used to create their kind of items.
 */
class MathExpressionFromXMLCreator
{
	public:
		/**
		 * @brief Access to the global instance
		 */
		static MathExpressionFromXMLCreator* the();

		/**
		 * @brief Create MathExpressions from xml
		 *
		 * Either from a QString or from a QDomNode
		 *
		 * @{
		 */
		MathExpression* fromXML( QString, QTreeItem* parent );
		MathExpression* fromXML( QDomNode, QTreeItem* parent );
		// @}

		/**
		 * @brief Return a list of all the elements creatable
		 *
		 * This asks all the registered creators for their list of elements and returns the combination
		 *
		 * @see elementsICanCreate
		 */
		QStringList availableElements() const;
		/**
		 * @brief Return a helptext for a certain type of element
		 *
		 * This can be used for GUI to easen the decision.
		 */
		QString helpText( QString ) const;

	protected:
		/**
		 * @brief default-constructor
		 *
		 * Registers itself at the global list of creators
		 */
		MathExpressionFromXMLCreator();
		/**
		 * @brief destructor
		 *
		 * Unregisters itself from the global list of creators
		 */
		virtual ~MathExpressionFromXMLCreator() {
			_creators.removeAll( this );
		}

		/**
		 * @brief Construct a single element from xml
		 */
		virtual MathExpression* singleElementFromXML( QDomNode, QTreeItem* pexp );

		/**
		 * @brief Return the string list of elements to create
		 */
		virtual QStringList elementsICanCreate() const;

		/**
		 * @brief Return the help text for a certain element-type
		 */
		virtual QString myHelpText( QString ) const;

	private:
		/**
		 * @brief No copying allowed.
		 */
		MathExpressionFromXMLCreator( MathExpressionFromXMLCreator& ) {}

		/**
		 * @internal
		 *
		 * @brief Stupid constructor to be used by the()
		 *
		 * This one doesn't register itself at the global list of creators,
		 * doing so inside of the() would result in an endless loop.
		 */
		MathExpressionFromXMLCreator( bool ) {}

		/**
		 * @brief List of all creators
		 */
		QList<MathExpressionFromXMLCreator*> _creators;
};

}; // NSCore

#endif // MATHEXPRESSION_H

