/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "diskiobuffers_bin.h"
#include "diskiobuffers_bin.moc"

#include <QtCore/QDebug>
#include <QtCore/QCoreApplication>
#include <QtCore/QBuffer>

namespace NSDiskIO {

	EPOSBinWriter::EPOSBinWriter( QObject* p )
		: AbstractWriter( p )
		, AbstractDiskIO<QDataStream>( QIODevice::WriteOnly | QIODevice::Append )
		, _strip( false )
	{
		setObjectName( "EPOSBinWriter" );
		config->addParameter( "Strip sampled data", QVariant::Bool, "If activated only the results will be saved to disk, the samples of the data will be stripped." );
		config->getParameter( "Strip sampled data" )->setValue( false );
		connect( config->getParameter( "Strip sampled data" ), SIGNAL( valueChanged( bool ) ), this, SLOT( stripChanged( bool ) ) );
	}
	EPOSBinWriter::~EPOSBinWriter() {
	}

	void EPOSBinWriter::stripChanged( bool n ) {
		_strip = n;
	}

	void EPOSBinWriter::setFile( QString filename ) {
		AbstractDiskIO<QDataStream>::setFile( filename );
		_file->write( "\%EPOSBIN3\%" );
	}

	void EPOSBinWriter::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		//qDebug() << "EPOSBinWriter::writeDataSet(" << &data << ")";
		if ( _stream == 0 )
			return;

		*_stream << data.channels()
		         << ( _strip?0:data.samples() )
		         << data.resulttree()
		         ;
		if ( ! _strip )
			for ( int i=0; i<data.channels(); ++i ) {
				double* ch = data.getRawChannel( i );
				for ( int j=0; j<data.samples(); ++j )
					*_stream << ch[ j ];
			}
	}

	EPOSBinReader::EPOSBinReader( QObject* p )
		: AbstractReader( p )
		, AbstractDiskIO<QDataStream>( QIODevice::ReadOnly )
		, _headertoskip( 0 )
	{
		setObjectName( "EPOSBinReader" );
	}
	EPOSBinReader::~EPOSBinReader() {
	}
	void EPOSBinReader::reset() {
		if ( _stream ) {
			_stream->device()->reset();
			_file->read( _headertoskip );
		}
	}
	void EPOSBinReader::setFile( QString file ) {
		AbstractDiskIO<QDataStream>::setFile( file );
		_headertoskip = 0;
		_fileversion = 0;
		if ( _file != 0 && _file->peek( 1 ) == "\%" ) {
			QByteArray tmp = _file->read( 1 );
			QByteArray r;

			do {
				r = _file->read( 1 );
				tmp.append( r );
			} while ( r != "\%" );

			_headertoskip = tmp.size();

			//if ( tmp == "\%EPOSBIN2\%" )
			//	qDebug() << "EPOSBinReader::setFile(" << file << ") seems to be a version 2 file.";
			if ( tmp == "\%EPOSBIN3\%" ) {
				qDebug() << "EPOSBinReader::setFile(" << file << ") found a version 3 file.";
				_fileversion = 1;
			}
		}
	}
	QString EPOSBinReader::filename() const {
		return AbstractDiskIO<QDataStream>::filename();
	}
	void EPOSBinReader::readSingleDataSet() {
		if ( _stream != 0 ) {

			if ( _fileversion == 0 ) {

				double samplinginterval, triggerdelay, horizontalposition, timestamp;
				int channels, samples;
				*_stream >> samplinginterval >> triggerdelay >> horizontalposition >> timestamp >> channels >> samples;

				NSCore::AcqisitionDataSet data( channels, samples );
				data.addResult( "samplinginterval", samplinginterval );
				data.addResult( "triggerdelay", triggerdelay );
				data.addResult( "horizontalposition", horizontalposition );
				data.addResult( "timestamp", timestamp );

				double fullscale, offset;
				for ( int i=0; i<channels; ++i ) {
					*_stream >> fullscale >> offset;
					data.addResult( QString( "channel %1/fullscale" ).arg( i ), fullscale );
					data.addResult( QString( "channel %1/offset" ).arg( i ), offset );

					double* ch = data.getRawChannel( i );
					for ( int j=0; j<samples; ++j ) {
						*_stream >> ch[ j ];
					}
				}
				data.updateMinMax();

				enqueue( data );

			} else if ( _fileversion == 1 ) {

				int channels, samples;
				*_stream >> channels >> samples;
				NSCore::AcqisitionDataSet data( channels, samples );
				*_stream >> data.resulttree();

				for ( int i=0; i<data.channels(); ++i ) {
					double* ch = data.getRawChannel( i );
					for ( int j=0; j<data.samples(); ++j ) {
						*_stream >> ch[ j ];
					}
				}
				data.updateMinMax();

				//data.debugPrint();
				enqueue( data );

			}
		}
	}



	PragReader::PragReader( QObject* p )
		: AbstractReader( p )
		, AbstractDiskIO<QDataStream>( QIODevice::ReadOnly )
	{
		setObjectName( "PragueReader" );
	}
	PragReader::~PragReader() {
	}
	void PragReader::setFile( QString file ) {
		AbstractDiskIO<QDataStream>::setFile( file );
	}
	QString PragReader::filename() const {
		return AbstractDiskIO<QDataStream>::filename();
	}
	void PragReader::readSingleDataSet() {
		if ( _stream != 0 ) {

			QVector<double> tdata;
			qint8 tmp = 0;
			while ( tmp == 0 && !_stream->atEnd() )
				*_stream >> tmp;

			while ( tmp != 0 && !_stream->atEnd() ) {
				tdata.push_back( tmp );
				*_stream >> tmp;
			}

			NSCore::AcqisitionDataSet data( 2/*channels*/, ( tdata.size() - 1 )/2 /*samples*/ );

			for ( int i=0; i<2; ++i ) {
				double *ch = data.getRawChannel( i );
				for ( int j=0; j<data.samples(); ++j ) {
					ch[ j ] = ( tdata[ j*2 + i ]-127 )/258;
				}
				//data.addResult( QString( "channel %1/fullscale" ).arg( i ), 1.0 );
			}
			data.updateMinMax();

			enqueue( data );
		}
	}

};

