/*
    Copyright 2007-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
                        Nicki Hinsche <nicvok@freenet.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "dopplerplugins.h"
#include <math.h>

TimeConstantAnalyzer::TimeConstantAnalyzer(  NSCore::QTreeItem* p ) : NSEvaluation::Analyzer(  p ) {
}

void TimeConstantAnalyzer::analyze(  NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<1/*data.channels()*/; ++i ) {
		//double max = data.channel( i )->result( "max" )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		int n_samples = data.samples();

		///results by linear regression
		double time_constant = lin_regression( max_at, data.getRawChannel( i ), n_samples );

		data.addResult( QString( "channel %1/time constant" ).arg( i ), time_constant );
		qDebug() << "time constant =" << time_constant;
	}
}

///lineare Regression _> aus y= a*exp(t) wird log(y)= C + t*x fit nach t, wobei x noch einfach die samples sind zB i=250ps

double TimeConstantAnalyzer::lin_regression( int x_max, double* ch, int n_0 ) const {

	int x_start = x_max + 20;     		  //sample position where regression starts

	int n = n_0 - x_max;                            //data.channels size -> how many samples to fit
	double kn = 1.0/n;                              //easier to calculate later

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=0; i<n; ++i){                     //or for(int i=x_max; i<data.channels(); ++i)
		sumX+= x_start + i;                     //summed samples
		sumY+= log( ch[ x_start + i ] );  //summed values of the samples
	}
	double sumXX = 0.0;
	double sumXY = 0.0;

	for(int i=0; i<n; ++i){
		sumXX+=(x_start + i)*(x_start + i);
		sumXY+=log( ch[ x_start + i ] )*(x_start + i);
	}

	double tau = (kn*sumXY - kn*sumX*kn*sumY)/(kn*sumXX-kn*sumX*kn*sumX);
	//double const = kn*sumY - tau*kn*sumX; 

	return tau;

}

IntegrationAnalyzerCF::IntegrationAnalyzerCF( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

	_config->addParameter( "low fraction point", 0.2 );
	_config[ "low fraction point" ]->setMin( 0.0 );
	_config[ "low fraction point" ]->setMax( 1.0 );
	_config->addParameter( "high fraction point", 0.9 );
	_config[ "high fraction point" ]->setMin( 0.0 );
	_config[ "high fraction point" ]->setMax( 1.0 );
	_config->addParameter( "expected mean value", 5.0);
	_config[ "expected mean value" ]->setMin( 0.0 );
	_config[ "expected mean value" ]->setMax( 100.0 );
	_config->addParameter( "expected deviation", 5.0 );
	_config[ "expected mean value" ]->setMin( 0.0 );
	_config[ "expected mean value" ]->setMax( 100.0 );

}

void IntegrationAnalyzerCF::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();

		int n_samples = data.samples();
		double _low_fraction_point = _config->getParameter( "low fraction point" )->toDouble();
		double _high_fraction_point = _config->getParameter( "high fraction point" )->toDouble();
		double _expec_mean = _config->getParameter( "expected mean value" )->toDouble();
		double _expec_deviation = _config->getParameter( "expected deviation" )->toDouble();


		///results
		double area_cf = check_cf( _low_fraction_point, _high_fraction_point, max_at, min_at, data.getRawChannel( i ), n_samples );
		double area_cfV2 = check_cfV2( _low_fraction_point, _high_fraction_point, max_at,  min_at, _expec_mean, _expec_deviation, data.getRawChannel( i ), n_samples );

		data.addResult( QString( "channel %1/cf" ).arg( i ), area_cf );
		data.addResult( QString( "channel %1/cf fullpulse" ).arg( i ), area_cfV2);

		qDebug() << " min(x,y), max(x,y): " << min << min_at << max << max_at;
		qDebug() << " Integral pulse after max with respect to CF: " << area_cfV2;
		qDebug() << " Integral only with respect to CF: " << area_cf;
	}
}



double IntegrationAnalyzerCF::check_cf( double lfp, double hfp, int x_max, int x_min, double* ch, int n_0 ) const {
		
	double max_val 	= ch[ x_max ];	//value of maximum
	double min_val 	= ch[ x_min ];	//value of minimum


	double scale 	= max_val-min_val;	//range between min and max

	int hcf_at=0;
	int lcf_at=0;
	double hcf=0;
	double lcf=0;

	//at which sample raises the pulse-value above the min CF before the maximum

	for(int i=x_max; i<n_0; i++){
		if ( ch[ i ] > (scale*hfp+min_val) ) {
				hcf = ch[ i ];	//high constant fraction
				hcf_at = i;
			}
	}

	//at which sample is the value below high CF after the maximum

	for(int i=0; i<x_max; i++){
		if ( ch[ i ] < (scale*lfp+min_val) ) {
				lcf = ch[ i ];	//low constant fraction
				lcf_at = i;
			}
	}


	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=lcf_at; i<hcf_at; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= i;			//summed samples
		sumY+= (ch[ i ]);		//summed values of the samples
	}

	double flaeche_wcf = (sumX*sumY)/1000000000; //normalization to get useful numbers with 10^9

	return flaeche_wcf;

}
double IntegrationAnalyzerCF::check_cfV2( double lfp, double hfp, int x_max, int x_min, double mean, double deviation, double* ch, int n_0 ) const {
		
	double max_val 	= ch[ x_max ];	//value of maximum
	double min_val 	= ch[ x_min ];	//value of minimum

	int x_start = x_max;
	int n = n_0 - x_max;	

	double scale 	= max_val-min_val;	//range between min and max

	int hcf_at=0;
	int lcf_at=0;
	double hcf=0;
	double lcf=0;

	//at which sample raises the pulse-value above the min CF before the maximum

	for(int i=x_max; i<n_0; i++){
		if ( ch[ i ] > (scale*hfp+min_val) ) {
				hcf = ch[ i ];	//high constant fraction
				hcf_at = i;
			}
	}

	//at which sample is the value below high CF after the maximum

	for(int i=0; i<x_max; i++){
		if ( ch[ i ] < (scale*lfp+min_val) ) {
				lcf = ch[ i ];	//low constant fraction
				lcf_at = i;
			}
	}


	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=lcf_at; i<hcf_at; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= i;			//summed samples
		sumY+= (ch[ i ]);	//summed values of the samples
	}

	double sumXX = 0.0;
	double sumYY = 0.0;

	double flaeche_wcf = (sumX*sumY)/1000000000; //normalization to get useful numbers with 10^9

	//only get well shaped pulses

	if(flaeche_wcf < (mean + deviation) && flaeche_wcf > (mean - deviation)){
		for(int i=0; i<n; ++i){
		sumXX+= x_start + i;
		sumYY+= (ch[ x_start + i ]);
	}
	}else {
		sumXX=NAN;
		sumYY=NAN;
	}

	double flaeche_wcf_V2 = sumYY/1000;


	return flaeche_wcf_V2;
	
}

IntegrationAnalyzer::IntegrationAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

}

void IntegrationAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		int n_samples = data.samples();

		///results
		double area = sumup( max_at, min_at, data.getRawChannel( i ), n_samples );
		
		data.addResult( QString( "channel %1/fullpulse" ).arg( i ), area );

		qDebug() << " min(x,y), max(x,y): " << min << min_at << max << max_at;
		qDebug() << " Integral pulse after max: " << area;
	}
}


double IntegrationAnalyzer::sumup( int x_max, int x_min, double* ch, int n_0 ) const {
		
	int x_start = x_max;			//sample position of maximum where summing starts
	int n = n_0 - x_max;			//data.channels size -> how many samples to fit

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=0; i<n; ++i){				//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= x_start + i;			//summed samples
		sumY+= (ch[ x_start + i ]);	//summed values of the samples
	}

	double flaeche = sumY/1000;

	return flaeche;

}

IntegrationAnalyzerCFspline::IntegrationAnalyzerCFspline( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

}

void IntegrationAnalyzerCFspline::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		int frac_at = data.result( QString( "channel %1/interpolated fraction point at" ).arg( i ) )->toInt();

		///results
		double area_cf_cf = cfarnold( frac_at, max_at, min_at, max, min, data.getRawChannel( i ), data.samples() );

		data.addResult( QString( "channel %1/cfcf" ).arg( i ), area_cf_cf );

	}
}

double IntegrationAnalyzerCFspline::cfarnold( double x_frac, int x_max, int x_min, double y_max, double y_min, double* ch, int n_0 ) const {
		
///intgeration of the area between the fraction point an the max. of the pulse

	int x_start = int( x_frac );
	int x_end = x_max;
	int n = n_0 - x_frac;			//data.channels size -> how many samples to fit

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=x_start; i<x_end; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= i;			//summed samples
		sumY+= (ch[ i ]);	//summed values of the samples
	}

	//double flaeche_fpm = (sumX*sumY)/1000000000; //no

///intgration of the area between the fraction point and the fraction point after the max. of the pulse

	double sumXX = 0.0;
	double sumYY = 0.0;

	int fp_after_max_at = 0;
	for(int i=x_end; i<n; i++){
		if ( ch[ i ] > ch[ x_start ] ) {
				fp_after_max_at = i;
			}
	}

	for(int i=x_start; i<fp_after_max_at; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumXX+= i;			//summed samples
		sumYY+= (ch[ i ]);	//summed values of the samples
	}

	
	double flaeche_fpfp = (sumXX*sumYY)/1000000000; //normalization to get useful numbers with 10^9


	return flaeche_fpfp;
	
}

IntegrationAnalyzerDiff::IntegrationAnalyzerDiff( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

	_config->addParameter( "Height difference per channel", 500.0 );
	_config[ "Height difference per channel" ]->setMin( 10.0 );
	_config[ "Height difference per channel" ]->setMax( 1000.0 );
}

void IntegrationAnalyzerDiff::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		int frac_at = data.result( QString( "channel %1/interpolated fraction point at" ).arg( i ) )->toInt();

		double _difference = _config->getParameter( "Height difference per channel" )->toDouble();

		///results
		double area_diff = diff( frac_at, _difference, max_at, min_at, max, min, data.getRawChannel( i ), data.samples() );

		data.addResult( QString( "channel %1/diff" ).arg( i ), area_diff );

	}
}

double IntegrationAnalyzerDiff::diff( int frac_at, double difference, int x_max, int x_min, double y_max, double y_min, double* ch, int n_0 ) const {

	///look for pile ups after the channel set by cf-plugin
	///take a look on the derivative of the pulse (digital p.u.r.)
		

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=frac_at; i<n_0; ++i){

		if( ((ch[i+1] - ch[i]) < difference) || ((ch[i+1] - ch[i]) > difference) ){
		sumX+= i;
		sumY+= (ch[ i ]);}
		else {
		qDebug() << " Height difference between conceeding channels " << (ch[i+1] - ch[i]);
	}
	}

	double flaeche_diff = (sumX*sumY)/1000000000; //normalization to get useful numbers with 10^9

	return flaeche_diff;
}






IntegrationAnalyzerZeroLine::IntegrationAnalyzerZeroLine( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

	_config->addParameter( "Integration till max (0), or full (1)", 0 );
	_config[ "Integration till max (0), or full (1)" ]->setMin( 0 );
	_config[ "Integration till max (0), or full (1)" ]->setMax( 1 );

}

void IntegrationAnalyzerZeroLine::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		int frac_at = data.result( QString( "channel %1/interpolated fraction point at" ).arg( i ) )->toInt();
		double zero = data.result( QString( "channel %1/zeroline" ).arg( i ) )->toDouble();

		int choose = _config->getParameter( "Integration till max (0), or full (1)" )->toInt();
		
		///results
		double area_zeroline = IntegrateZeroline( frac_at, max_at, min_at, max, min, zero, choose, data.getRawChannel( i ), data.samples() );

		data.addResult( QString( "channel %1/zeroline integral" ).arg( i ), area_zeroline );

	}
}

double IntegrationAnalyzerZeroLine::IntegrateZeroline( double x_frac, int x_max, int x_min, double y_max, double y_min, double zeroline, int select, double* ch, int n_0 ) const {
		
///intgration of the area between the fraction point and the max./full of the pulse
///heigth up the zeroline

	int x_start = int( x_frac );
	

	/*if ( select == 0 || _extremum_at_parameter == 0 ) {
			qDebug() << " Integration of the pulse until the maximum";
			int x_end = x_max
			return;
		} else {qDebug() << " Integration of the full pulse";
			int x_end = */
			

	//int n	= n_0;

///intgration of the area between the fraction point and the fraction point after the max. of the pulse

	double sumXX = 0.0;
	double sumYY = 0.0;

	for(int i=x_start; i<n_0; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumXX+= i;			//summed samples
		sumYY+= (ch[ i ]-zeroline);	//summed values of the samples, corrected with the zeroline
	}

	
	//double flaeche_zero = (sumXX*sumYY)


	//return flaeche_zero;
	return n_0;
}















IntegrationAnalyzerAnalytic::IntegrationAnalyzerAnalytic( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {


	_config->addParameter( "1st constant fraction", 0.2 );
	_config[ "1st constant fraction" ]->setMin( 0.0 );
	_config[ "1st constant fraction" ]->setMax( 0.5 );
	_config->addParameter( "2nd constant fraction", 0.8 );
	_config[ "2nd constant fraction" ]->setMin( 0.5 );
	_config[ "2nd constant fraction" ]->setMax( 1.0 );
	_config->addParameter( "negative length of the calculated pulse", 1.0);
	_config[ "negative length of the calculated pulse" ]->setMin( 0.0 );
	_config[ "negative length of the calculated pulse" ]->setMax( 500.0 );

}

void IntegrationAnalyzerAnalytic::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<data.channels(); ++i ) {
		double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		double zero = data.result( QString( "channel %1/zeroline" ).arg( i ) )->toDouble();


		//int n_samples = data.samples();
		double _low_fraction_point =  _config->getParameter( "1st constant fraction" )->toDouble();
		double _high_fraction_point =  _config->getParameter( "2nd constant fraction" )->toDouble();
		//int _pulse_length = _config->getParameter( "negative length of the calculated pulse" )->toInt();


		///results
		double area_analytic = IntegrateAnalytic( _low_fraction_point, _high_fraction_point,  max_at, min_at, max, min, zero, data.getRawChannel( i ), data.samples() );



		data.addResult( QString( "channel %1/analytic integral" ).arg( i ), area_analytic );

	}
}

double IntegrationAnalyzerAnalytic::IntegrateAnalytic( double lfp, double hfp, int x_max, int x_min, double y_max, double y_min, double zeroline, double* ch, int n_0 ) const {
	

	double max_val 	= ch[ x_max ];	//value of maximum
	double min_val  = zeroline;	//value of zeroline



	double scale 	= max_val-min_val;	//range between min and max

	double hcf =0;
	double lcf =0;
	int hcf_at=0;
	int lcf_at=0;

	//at which sample raises the pulse-value above the min CF before the maximum

	for(int i=x_max; i<n_0; i++){
		if ( ch[ i ] > (scale*hfp+min_val) ) {
				hcf = ch[ i ];	//high constant fraction
				hcf_at = i;
			}
	}

	//at which sample is the value below high CF after the maximum

	for(int i=0; i<x_max; i++){
		if ( ch[ i ] < (scale*lfp+min_val) ) {
				lcf = ch[ i ];	//low constant fraction
				lcf_at = i;
			}
	}

///find an expression for the zeroline	

	int zeroline_reg_start_at = 0;
	int x_start = 0;
	//int zeroline_reg_end_at = lcf_at-20;

	int n = lcf_at-20; 	                         //how many samples to fit
	double kn = 1.0/n;                              //easier to calculate later

	double sumX_reg_zero = 0.0;
	double sumY_reg_zero = 0.0;

	for(int i=0; i<n; ++i){                     		    
		sumX_reg_zero+= zeroline_reg_start_at + i;         //summed samples
		sumX_reg_zero+= ch[ zeroline_reg_start_at + i ];  //summed values of the samples
	}

	double sumXX_reg_zero = 0.0;
	double sumXY_reg_zero = 0.0;

	for(int i=0; i<n; ++i){
		sumXX_reg_zero+=(x_start + i)*(x_start + i);
		sumXY_reg_zero+=log( ch[ x_start + i ] )*(x_start + i);
	}


	///sloope of the zeroline

	double m_zeroline = (kn*sumXY_reg_zero - kn*sumX_reg_zero*kn*sumY_reg_zero)/(kn*sumXX_reg_zero-kn*sumX_reg_zero*kn*sumX_reg_zero);

	
	return m_zeroline;

}








TimeConstantIntegrationAnalyzer::TimeConstantIntegrationAnalyzer(  NSCore::QTreeItem* p ) : NSEvaluation::Analyzer(  p ) {


	_config->addParameter( "lenght of the evaluated pulse", 100 );
	_config[ "lenght of the evaluated pulse" ]->setMin( 10 );
	_config[ "lenght of the evaluated pulse" ]->setMax( 300 );
	_config->addParameter( "distance from cf at (for regression)", 20 );
	_config[ "distance from cf at (for regression)" ]->setMin( 0 );
	_config[ "distance from cf at (for regression)" ]->setMax( 500 );
	_config->addParameter( "distance from cf at (for integration)", 0 );
	_config[ "distance from cf at (for integration)" ]->setMin( 0 );
	_config[ "distance from cf at (for integration)" ]->setMax( 500 );

}

void TimeConstantIntegrationAnalyzer::analyze(  NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {
		//double max = data.result( QString( "channel %1/max" ).arg( i ) )->toDouble();
		int max_at = data.result( QString( "channel %1/max at" ).arg( i ) )->toInt();
		//double min = data.result( QString( "channel %1/min" ).arg( i ) )->toDouble();
		//int min_at = data.result( QString( "channel %1/min at" ).arg( i ) )->toInt();
		double zero = data.result( QString( "channel %1/zeroline" ).arg( i ) )->toDouble();
		int frac_at = data.result( QString( "channel %1/interpolated fraction point at" ).arg( i ) )->toInt();
		int n_samples = data.samples();


		int _dist_reg = _config->getParameter( "distance from cf at (for regression)" )->toInt();
		int _dist_int = _config->getParameter( "distance from cf at (for integration)" )->toInt();
		int _length = _config->getParameter( "lenght of the evaluated pulse" )->toInt();
		

		///results by linear regression
		double analytic_int = lin_regression_integrate( frac_at, zero, max_at, _dist_reg, _dist_int, _length, data.getRawChannel( i ), n_samples );

		data.addResult( QString( "channel %1/analytic integral" ).arg( i ), analytic_int );
		qDebug() << "distance regression =" << _dist_reg;
		qDebug() << "distance integral =" << _dist_int;
		qDebug() << "analytic integral =" << analytic_int;
	}
}


///lineare Regression _> aus y= a*exp(t) wird log(y)= C + t*x fit nach t, wobei x noch einfach die samples sind zB i=250ps
///afterwards the pulse will be evaluated (integrated) analytically in the given borders



double TimeConstantIntegrationAnalyzer::lin_regression_integrate( int xfrac_at, double xzero, int xmax_at, int x_dist_reg, int x_dist_int, int x_length, double* ch, int n_0 ) const {

	int x_start = xfrac_at + x_dist_reg;              //sample position where regression of the pulse starts

	int n = x_length;     			         //data.channels size -> how many samples to fit
	double kn = 1.0/n;                              //easier to calculate later

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=0; i<n; ++i){                    	  //or for(int i=x_max; i<data.channels(); ++i)
		sumX+= x_start + i;                      //summed samples
		sumY+= log( ch[ x_start + i ]- xzero);  //summed zero-corrected values of the samples
	}
	double sumXX = 0.0;
	double sumXY = 0.0;

	for(int i=0; i<n; ++i){
		sumXX+=(x_start + i)*(x_start + i);
		sumXY+=log( ch[ x_start + i ] - xzero)*(x_start + i);
	}

	double tau = (kn*sumXY - kn*sumX*kn*sumY)/(kn*sumXX-kn*sumX*kn*sumX);
	//double const = kn*sumY - tau*kn*sumX; 


	double analytic_int=(exp(-tau*(xfrac_at+x_dist_reg+x_length))-exp(-tau*(xfrac_at+x_dist_reg)));


	return analytic_int
;

}

