/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

/**
 * \file
 * This file contains the public interfaces for plugins
 */

#ifndef PLUGINS_INTERFACE_H
#define PLUGINS_INTERFACE_H

#include <QtCore/QString>
#include <QtCore/QList>

#include <datatypes.h>
#include <namedparametertreeitem.h>

namespace NSEvaluation {

	/// Base-class for plugins in Evaluation
	/**
	 * Provides a NamedParameterTreeItem _config and access to it via
	 * NamedParameterTreeItemPointer config();
	 */
	class Plugin
	{
		public:
			Plugin( NSCore::QTreeItem* p );
			virtual ~Plugin() {}

			/// Access to the configuration
			NSCore::NamedParameterTreeItemPointer config();

			/// Return the type of this plugin
			virtual QString type() const =0;
			virtual QString help() const { return "<qt><i>Sorry, no help entered.</i></qt>"; }

		protected:
			/**
			 * Use _config to store and access your configuration parameters.
			 *
			 * These will be made accessible by the GUIs
			 */
			NSCore::NamedParameterTreeItemPointer _config;
	};

	/// The base-class for Analyzers.
	/**
	 * Analyzers do the real work, the results should be stored in
	 * NSCore::ChannelDataSet::addResult().
	 *
	 * @see analyze
	 */
	class Analyzer : public Plugin
	{
		public:
			Analyzer( NSCore::QTreeItem* p ) : Plugin( p )
			{ }
			/// The evaluation function
			/**
			 * This function needs to be implemented and is called to do the
			 * real evaluation.
			 * The results should be added to the data, either per dataset or
			 * per channel, depending on the analysation.
			 *
			 * @param data The dataset to analyze
			 */
			virtual void analyze( NSCore::AcqisitionDataSet& data ) =0;
	};

	/// The base-class for Modifiers.
	/**
	 * Modifiers return a new dataset which is ( should be? ) derived from the
	 * given set.
	 *
	 * @see modify
	 */
	class Modifier : public Plugin
	{
		public:
			Modifier( NSCore::QTreeItem* p ) : Plugin( p )
			{ }

			/**
			 * @brief Hook for preparations
			 *
			 * Implement this function if you need to do some things before the
			 * real new generation is created.
			 *
			 * For example you can analyze the data to check whether your
			 * prerequisites are fulfilled. And you you set/change channels and
			 * samples to the values you like. That way your new generation (if
			 * there is one to be created) will have the parameters you want.
			 */
			virtual void prepare( NSCore::AcqisitionDataSet& data ) {
				prefix = type();
				samples = data.samples();
				channels = data.channels();
			}

			/**
			 * @brief The modifier function
			 * This function is called to do the real modifications of the data.
			 * A new instance of data is created before this function gets called.
			 *
			 * If you want to check whether a new generation was created, just
			 * compare oldgeneration with data.currentGenerationNumber().
			 */
			virtual void modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) =0;

			/**
			 * @brief The prefix to add to the result-tree
			 */
			QString prefix;
			/**
			 * @brief Number of samples to create in the new generation
			 */
			int samples;
			/**
			 * @brief Number of channels to create in the new generation
			 */
			int channels;
	};

	/// The base-class for Filters.
	/**
	 * Filters check the data for any kind of prerequisites and return true or
	 * false.
	 *
	 * @see filter
	 */
	class Filter : public Plugin
	{
		public:
			Filter( NSCore::QTreeItem* p ) : Plugin( p )
			{ }
			/// The filter function
			/**
			 * This function analyzes the data and decides wether the evaluation should
			 * go on.
			 */
			virtual bool filter( NSCore::AcqisitionDataSet& ) =0;
	};


	/// PluginFactories public client
	/**
	 * Subclasses of this class register themselve at the PluginFactory to create
	 * new plugins.
	 *
	 * You should instantiate an object of your class in your lib so the client
	 * registers at loading...
	 */
	class PluginFactory_Client {
		public:
			PluginFactory_Client();
			virtual ~PluginFactory_Client() {}

			/// What can be created by this factory.
			virtual QStringList canCreate() =0;
			/// Create a named item with the given parent.
			virtual Plugin* create( QString, NSCore::QTreeItem* ) =0;
			/// Name the types of the plugins
			/**
			 * Has to return either "Analyzer", "Modifier" or "Filter"
			 */
			virtual QString type( QString plugin ) =0;

			virtual QString help( QString plugin ) { return "<qt><b>To be implemented!</b><br>The programmer responsible for this plugin hasn't yet added documentation for it.</qt>"; }
			virtual QStringList categories( QString plugin ) =0;

			virtual QStringList results( QString plugin ) { return QStringList(); }
	};

};

#endif // PLUGINS_INTERFACE_H

