/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "qtreeitemmodeldelegate.h"
#include "qtreeitemmodeldelegate.moc"

#include <QtCore/QDebug>
#include <QtCore/QModelIndex>
#include <QtCore/QEvent>
#include <QtCore/QUrl>
#include <QtGui/QMouseEvent>
#include <QtGui/QPainter>
#include <QtGui/QItemEditorCreatorBase>
#include <QtGui/QDoubleSpinBox>

#include <qtreeitemmodel.h>

class DoubleCreator : public QItemEditorCreatorBase
{
	public:
		QWidget* createWidget( QWidget* p ) const {
			QDoubleSpinBox* tmp = new QDoubleSpinBox( p );
			tmp->setMinimum( -10000.0 );
			tmp->setMaximum( 10000.0 );
			tmp->setDecimals( 5 );
			return tmp;
		}
		QByteArray valuePropertyName() const { return "value"; }
};

QTreeItemModelDelegate::QTreeItemModelDelegate( QObject* p ) : QItemDelegate( p ), _gui( 0 ) {
	init();
}
QTreeItemModelDelegate::QTreeItemModelDelegate( NSCore::GuiServer_Interface* gui, QObject* p )
	: QItemDelegate( p )
	, _gui( gui )
{
	init();
}

void QTreeItemModelDelegate::init() {
	QItemEditorFactory* factory = new QItemEditorFactory();
	factory->registerEditor( QVariant::Double, new DoubleCreator() );
	setItemEditorFactory( factory );
}

QSize QTreeItemModelDelegate::sizeHint( const QStyleOptionViewItem& option, const QModelIndex& index ) const {
	//return QItemDelegate::sizeHint( option, index ) + QSize( 0, 2 );
	QSize hint;
	QVariant deco = index.data( Qt::DecorationRole );
	QVariant text = index.data( Qt::DisplayRole );
	if ( deco.type() == QVariant::String || deco.type() == QVariant::Icon ) {
		hint = hint.expandedTo( QSize( 20, 20 ) );
	}
	QFontMetrics metrics( option.font );
	hint = hint.expandedTo( metrics.boundingRect( text.toString() ).size() );
	//qDebug() << "QTreeItemModelDelegate::sizeHint() will return" << hint;
	return hint;
}

void QTreeItemModelDelegate::paint( QPainter* p, const QStyleOptionViewItem& option, const QModelIndex& index ) const {
	if ( !index.isValid() )
		return;

	if ( index.column() == 1 ) {
		if ( index.data( Qt::DisplayRole ).type() == QVariant::Url ) {
			drawBackground( p, option, index );
			drawDisplay( p, option, option.rect, index.data( Qt::DisplayRole ).toUrl().toString() );
		} else
			QItemDelegate::paint( p, option, index );
		return;
	}

	drawBackground( p, option, index );

	if ( index.column() == 0 ) {
		QIcon icon;
		QVariant d = index.data( Qt::DecorationRole );
		if ( d.type() == QVariant::Icon )
			icon = d.value<QIcon>();
		if ( d.type() == QVariant::String ) {
			icon = QIcon( d.toString() );
		}
		if ( !icon.isNull() ) {
			int s = sizeHint( option, index ).height();
			QRect rect = option.rect;
			rect.setWidth( s );
			drawDecoration( p, option, rect, icon.pixmap( s, s ) );
		}

		drawDisplay( p, option, option.rect.adjusted( ( icon.isNull() ? 0 : sizeHint( option, index ).height() ), 0, 0, 0 ), index.data( Qt::DisplayRole ).toString() );
		return;
	}

	QStringList actions;
	{
		QVariant data = index.model()->data( index, Qt::DisplayRole );
		if ( data.type() == QVariant::StringList )
			actions = data.toStringList();
	}

	if ( actions.size() > 0 ) {
		float w = option.rect.width();
		w = w/actions.size();
		QRectF rect = option.rect;
		rect.setWidth( w );

		QLinearGradient grad( rect.topLeft(), rect.bottomLeft() );
		QColor base = option.palette.base().color();
		if ( option.state & QStyle::State_Selected ) {
			base = option.palette.highlight().color();
			p->setPen( option.palette.highlightedText().color() );
		}
		grad.setColorAt( 0.1, base );
		grad.setColorAt( 0.3, base.dark( 125 ) );
		grad.setColorAt( 0.9, base );

		for ( int i=0; i<actions.size(); ++i ) {
			p->fillRect( rect.adjusted( 0,0, -1,-1 ), QBrush( grad ) );
			p->drawText( rect, Qt::AlignCenter /*option.displayAlignment*/, actions[ i ] );
			rect.translate( w, 0 );
		}
	}

}

bool QTreeItemModelDelegate::editorEvent( QEvent* ev, QAbstractItemModel* model, const QStyleOptionViewItem& option, const QModelIndex& index ) {
	if ( !index.isValid() || index.column() != 2 )
		return QItemDelegate::editorEvent( ev, model, option, index );

	//qDebug() << "TreeItemModelDelegate::editorEvent(" << ev << "," << model << "," << option << "," << index << ")";

	if ( ev->type() != QEvent::MouseButtonRelease )
		return false;

	QMouseEvent* mev = dynamic_cast<QMouseEvent*>( ev );

	QStringList actions;
	{
		QVariant data = index.model()->data( index, Qt::DisplayRole );
		if ( data.type() == QVariant::StringList )
			actions = data.toStringList();
	}

	if ( actions.size() > 0 ) {

		float w = option.rect.width();
		w = w/actions.size();
		QRectF rect = option.rect;
		rect.setWidth( w );

		for ( int i=0; i<actions.size(); ++i ) {
			if ( rect.intersects( QRect( mev->pos(), mev->pos() ) ) ) {
				//qDebug() << QString( "Selected action (%1) is \"%2\"" ).arg( i ).arg( actions[ i ] );

				NSCore::QTreeItemNode* node = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
				NSCore::QTreeItem* item = dynamic_cast<NSCore::QTreeItem*>( node->object() );
				if ( item ) {
					//qDebug() << " Trying to execute the action on" << item;
					item->guiExecAction( actions[ i ], _gui );
				}
			}
			rect.translate( w, 0 );
		}
	}

	return false;
}

