/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_FFTW

#include "fftw.h"

#include <cmath>
#include <QtCore/QDebug>

class FFTW_CleanUp
{
	private:
		FFTW_CleanUp() : refs( 0 ) {}
		~FFTW_CleanUp() {}

		int refs;

	public:
		static FFTW_CleanUp* the() {
			static FFTW_CleanUp* me = new FFTW_CleanUp();
			return me;
		}

		void cleanupregister() {
			refs++;
		}

		void cleanupunregister() {
			refs--;
			if ( refs <= 0 ) {
				qDebug() << "FFTW_CleanUp: refs reached zero, calling fftw_cleanup()";
				fftw_cleanup();
			}
		}
};

DFT_R2HC::DFT_R2HC( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ), _samples( 0 ), _p( 0 ), _in( 0 ), _out( 0 ) {
	FFTW_CleanUp::the()->cleanupregister();
}
DFT_R2HC::~DFT_R2HC() {
	fftw_destroy_plan( _p );
	delete[] _in;
	delete[] _out;
	FFTW_CleanUp::the()->cleanupunregister();
}

void DFT_R2HC::prepare( NSCore::AcqisitionDataSet& data ) {
	if ( data.samples() != _samples ) {
		_samples = data.samples();
		fftw_destroy_plan( _p );
		delete[] _in;
		delete[] _out;
		_in = new double[ _samples ];
		_out = new double[ _samples ];
		_p = fftw_plan_r2r_1d( _samples, _in, _out, FFTW_R2HC, FFTW_MEASURE | FFTW_PRESERVE_INPUT );
	}
	Modifier::prepare( data );
}

void DFT_R2HC::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double *_in = data.getRawChannel( i, oldgeneration );
		if ( oldgeneration == data.currentGenerationNumber() ) {
			_in = new double[ data.samples() ];
			memcpy( _in, data.getRawChannel( i, oldgeneration ), sizeof( double ) * data.samples() );
		}
		fftw_execute_r2r( _p, _in, data.getRawChannel( i ) );

		double *tmp = data.getRawChannel( i );
		double n = sqrt( data.samples() );
		for ( int j=0; j<data.samples(); ++j )
			tmp[ j ] /= n;
	}
	data.updateMinMax();
	//data.debugPrint();
}

DFT_NiceAnalysis::DFT_NiceAnalysis( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ), _samples( 0 ), _p( 0 ), _in( 0 ), _out( 0 ) {
	FFTW_CleanUp::the()->cleanupregister();
}
DFT_NiceAnalysis::~DFT_NiceAnalysis() {
	fftw_destroy_plan( _p );
	delete[] _in;
	delete[] _out;
	FFTW_CleanUp::the()->cleanupunregister();
}

void DFT_NiceAnalysis::prepare( NSCore::AcqisitionDataSet& data ) {
	Modifier::prepare( data );
	channels = data.channels() * 2;
	samples = data.samples() / 2;
	if ( data.samples() != _samples ) {
		_samples = data.samples();
		fftw_destroy_plan( _p );
		delete[] _in;
		delete[] _out;
		_in = new double[ _samples ];
		_out = new double[ _samples ];
		_p = fftw_plan_r2r_1d( _samples, _in, _out, FFTW_R2HC, FFTW_MEASURE | FFTW_PRESERVE_INPUT );
	}
}

void DFT_NiceAnalysis::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	//qDebug() << "DFT_NiceAnalysis::modify()";

	for ( int i=0; i<data.channels( oldgeneration ); ++i )
		fftw_execute_r2r( _p, data.getRawChannel( i, oldgeneration ), data.getRawChannel( i*2 ) );

	for ( int i=0; i<data.channels(); i+=2 ) {
		double *tmp = data.getRawChannel( i );
		double n = sqrt( data.samples( oldgeneration ) );
		for ( int j=0; j<data.samples(); ++j )
			tmp[ j ] = 10 * log( pow( tmp[ j ]/n, 2 ) );

		data.addResult( QString( "channel %1/fullscale" ).arg( i ), 250.0 );
		data.addResult( QString( "channel %1/offset" ).arg( i ), 100.0 );
	}

	for ( int i=1; i<data.channels(); i+=2 ) {
		double *tmp = data.getRawChannel( i );
		double n;
		for ( int j=0; j<data.samples()/2; ++j ) {
			n = tmp[ j ];
			tmp[ j ] = tmp[ data.samples() - j ];
			tmp[ data.samples() - j ] = n;
		}

		data.addResult( QString( "channel %1/fullscale" ).arg( i ), 50.0 );
		data.addResult( QString( "channel %1/offset" ).arg( i ), 0.0 );
	}

	data.updateMinMax();
	//data.debugPrint();
}


DFT_HC2R::DFT_HC2R( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ), _samples( 0 ), _p( 0 ), _in( 0 ), _out( 0 ) {
	FFTW_CleanUp::the()->cleanupregister();
}
DFT_HC2R::~DFT_HC2R() {
	fftw_destroy_plan( _p );
	delete[] _in;
	delete[] _out;
	FFTW_CleanUp::the()->cleanupunregister();
}

void DFT_HC2R::prepare( NSCore::AcqisitionDataSet& data ) {
	if ( data.samples() != _samples ) {
		_samples = data.samples();
		fftw_destroy_plan( _p );
		delete[] _in;
		delete[] _out;
		_in = new double[ _samples ];
		_out = new double[ _samples ];
		_p = fftw_plan_r2r_1d( _samples, _in, _out, FFTW_HC2R, FFTW_MEASURE | FFTW_PRESERVE_INPUT );
	}
	Modifier::prepare( data );
}

void DFT_HC2R::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double *_in = data.getRawChannel( i, oldgeneration );
		if ( oldgeneration == data.currentGenerationNumber() ) {
			_in = new double[ data.samples() ];
			memcpy( _in, data.getRawChannel( i, oldgeneration ), sizeof( double ) * data.samples() );
		}
		fftw_execute_r2r( _p, _in, data.getRawChannel( i ) );

		double *tmp = data.getRawChannel( i );
		double n = sqrt( data.samples() );
		for ( int j=0; j<data.samples(); ++j )
			tmp[ j ] /= n;
	}
	data.updateMinMax();
	//data.debugPrint();
}

HC_LowPass::HC_LowPass( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "maximum frequency", 0.5 );
	_config[ "maximum frequency" ]->setRange( 0.0, 1.0 );
}

void HC_LowPass::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	double freq = _config[ "maximum frequency" ]->toDouble();
	int skip = data.samples()/2 * freq;

	if ( oldgeneration != data.currentGenerationNumber() )
		memcpy( data.getRawData(), data.getRawData( oldgeneration ), data.samples() * data.channels() * sizeof( double ) );

	for ( int i=0; i<data.channels(); ++i ) {
		double* tmp = data.getRawChannel( i );
		for ( int j=skip ; j<data.samples()-skip; ++j )
			tmp[ j ] = 0.0;
	}
	data.updateMinMax();
}

#endif
