/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ACQIRIS_THREAD_H
#define ACQIRIS_THREAD_H

#include <QtCore/QThread>
#include <QtCore/QQueue>
#include <QtCore/QMultiMap>
#include <QtCore/QSemaphore>
#include <QtCore/QReadWriteLock>
#include <QtCore/QTime>

#include <datatypes.h>

#include <namedparametertreeitem.h>

#ifdef HAVE_ACQIRIS
#include "AcqirisD1Import.h"
#endif

#include <cmath>

namespace NSAcqiris {

	void errorMessage( const quint32 _inst, const qint32 status, const QString& comment="" ) {
		if ( status != 0 ) {
			ViChar msg[ 500 ];
			AcqrsD1_errorMessage( _inst, status, msg );
			qDebug() << QString( " Last call %1returned: %2" ).arg( comment.isEmpty()?"":"( "+comment+" ) " ).arg( msg );
		}
	}

	struct DigitizerChannelData {
		ViAddr dataArray;
		AqDataDescriptor* dataDesc;
		ViAddr segDescArray;
	};
	class DigitizerData {
	public:
		int channels;
		int segments;
		int fullDataSize;
		int dataBytes;

		QVector<DigitizerChannelData> data;

		bool filter;
		int filter_min_min, filter_min_max, filter_max_min, filter_max_max;

	private:
		friend class DigitizerDataPool;

		DigitizerData( int ch, int seg, int dS, int dB ) : channels( ch ), segments( seg ), fullDataSize( dS ), dataBytes( dB ), filter( false ) {
			for ( int i=0; i<channels; ++i ) {
				DigitizerChannelData tmp;
				tmp.dataArray = new char[ dS ];
				tmp.segDescArray = new AqSegmentDescriptor[ segments ];
				tmp.dataDesc = new AqDataDescriptor;
				data.push_back( tmp );
			}
		}
		~DigitizerData() {
			for ( int i=0; i<channels; ++i ) {
				DigitizerChannelData tmp = data.at( i );
				delete[] ( char* )tmp.dataArray;
				delete[] static_cast<AqSegmentDescriptor*>( tmp.segDescArray );
				delete tmp.dataDesc;
			}
		}
	};
	class DigitizerDataPool {
	private:
		DigitizerDataPool()
			: _poolLock( new QReadWriteLock() )
		{
			//dataPoolHits = dataPoolMissed = overallDataSets = 0;
		}
		~DigitizerDataPool() {
			//qDebug() << " DataPool was hit" << dataPoolHits << "times and missed" << dataPoolMissed << "times.\n Over all there where" << overallDataSets << "datasets needed.\n";
			delete _poolLock;
		}

		/// Pool
		QMultiMap<int, DigitizerData*> _pool;
		QReadWriteLock* _poolLock;
		QSemaphore _poolSemaphore;

		/// Housekeeping
		int _last_used_size;

		/// Statistics
		//int dataPoolHits, dataPoolMissed, overallDataSets;

	public:
		static DigitizerDataPool* the() {
			static DigitizerDataPool* me = 0;
			if ( ! me ) {
				me = new DigitizerDataPool();
				me->_poolSemaphore.release( 1000 );
			}
			return me;
		}

		DigitizerData* getData( int ch, int seg, int dS, int dB ) {
			DigitizerData* ret = 0;
			_poolSemaphore.acquire();
			_poolLock->lockForWrite();
			if ( _pool.count( dS ) > 0 ) {
				ret = _pool.take( dS );
				_last_used_size = dS;
				//dataPoolHits++;
			} else {
				ret = new DigitizerData( ch, seg, dS, dB );
				//dataPoolMissed++;
			}
			//overallDataSets++;
			_poolLock->unlock();
			return ret;
		}
		void deleteData( DigitizerData* data ) {
			_poolLock->lockForWrite();
			_pool.insert( data->fullDataSize, data );
			_poolSemaphore.release();
			_poolLock->unlock();
		}

		void houseKeeping( int size_to_preserve, int amount_to_preserve = 10 ) {
			_poolLock->lockForWrite();
			foreach( int size, _pool.keys() )
				if ( size != size_to_preserve )
					while ( _pool.count( size ) > amount_to_preserve )
						delete _pool.take( size );
			_poolLock->unlock();
		}
	};

	class DigitizerThread : public QThread
	{
		Q_OBJECT

		quint32 _inst;

		bool _run, _quit, _single, _verticalChanged, _horizontalChanged, _triggerChanged, _scheduleCalibration, _filterChanged;

		qint32 _channels, _externaltriggers;

		NSCore::NamedParameterTreeItem* _config;

		QQueue<DigitizerData*> _data;
		QMutex* _datamutex;

		int _last_used_size;
		QTime _last_cleanup;

		QTime _temperatur_time;
		NSCore::NamedParameterTreeItem* _temperatur_param;

		qint32 _datatype_to_read;

		QTime _statistics_time;
		NSCore::NamedParameterTreeItem* _statistics_cps_param;
		double _statistics_cps;

		NSCore::NamedParameterTreeItem* _acquisitionscounter;
		NSCore::NamedParameterTreeItem* _acquisitionscounter_max;

		QString _devicename;
		Digitizer* _digitizer;

		public:
			DigitizerThread( ViSession s, NSCore::NamedParameterTreeItem* config, Digitizer* p )
				: QThread( p )
				, _inst( s )
				, _run( false )
				, _quit( false )
				, _single( false )
				, _verticalChanged( true )
				, _horizontalChanged( true )
				, _triggerChanged( true )
				, _filterChanged( true )
				, _config( config )
				, _datamutex( new QMutex )
				, _datatype_to_read( 0 )
				, _statistics_cps( 0.0 )
				, _digitizer( p )
			{
				qDebug() << "DigitizerThread::DigitizerThread(" << s << "," << p << ")";
				errorMessage( AcqrsD1_getNbrChannels( _inst, ( ViInt32* )&_channels ), "AcqrsD1_getNbrChannels" );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrExternalTriggers", &_externaltriggers ), "getExternalTriggers" );

				char* options = new char[ 64 ];
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"Options", options ), "AcqrsD1_getInstrumentInfo" );
				qDebug() << " The digitizer seems to have the following options installed:" << options;
				qint32 tmp;
				errorMessage( AcqrsD1_getInstrumentData( _inst, options, ( ViInt32* )&tmp, ( ViInt32* )&tmp, ( ViInt32* )&tmp ), "AcqrsD1_getInstrumentData" );
				_devicename = options;
				qDebug() << " Device name is" << _devicename;
				delete[] options;
				options = 0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"MaxSamplesPerChannel", &tmp ) );
				qDebug() << " Maximum number of samples per channel is:" << tmp;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrADCBits", &tmp ) );
				qDebug() << " Number of ADC-bits:" << tmp;
				if ( tmp > 8 )
					_datatype_to_read = 1;

				qDebug() << endl;

				_config->addParameter( "Stop after N acquisitions", QVariant( 0 ) );
				_acquisitionscounter_max = _config->getParameter( "Stop after N acquisitions" );
				_acquisitionscounter_max->setRange( 0, int( 10E12 ) );
				_config->addParameter( "Acquisitions Counter", QVariant::ULongLong );
				_acquisitionscounter = _config->getParameter( "Acquisitions Counter" );
				_acquisitionscounter->setGuiHint( NSCore::NamedParameterTreeItem::Counter );

				for ( int i=1; i<=_channels; ++i ) {
					_config->addParameter( QString( "Channel %1" ).arg( i ), QVariant::Invalid );
					NSCore::NamedParameterTreeItem* tmp = _config->getParameter( QString( "Channel %1" ).arg( i ) );
					tmp->addParameter( "fullscale", "5V" ); // TODO enum
					tmp->addParameter( "offset", 0.0 );
					tmp->getParameter( "offset" )->setRange( -5.0, 5.0 );
					tmp->addParameter( "coupling", 3, "0 = Ground, 1 = DC 1MOhm, 2 = AC 1MOhm, 3 = DC 50Ohm, 4 = AC 50Ohm (DC440: 3=Std-In 4=HF-In)" ); // TODO enum
					tmp->getParameter( "coupling" )->setRange( 0,4 );
					tmp->addParameter( "bandwidth", QVariant( 0 ), "0 = No Bandlimit, 1 = 25MHz, 2 = 700MHz, 3 = 200MHz, 4 = 20MHz, 5 = 35MHz" ); // TODO enum
					tmp->getParameter( "bandwidth" )->setRange( 0,5 );
					connect( tmp, SIGNAL( valueChanged() ), this, SLOT( verticalChanged() ) );
				}
				for ( int i=-1; i>=-_externaltriggers; --i ) {
					_config->addParameter( QString( "Channel %1" ).arg( i ), QVariant::Invalid );
					NSCore::NamedParameterTreeItem* tmp = _config->getParameter( QString( "Channel %1" ).arg( i ) );
					tmp->addParameter( "fullscale", "5V" ); // TODO enum
					tmp->addParameter( "offset", 0.0 );
					tmp->getParameter( "offset" )->setRange( -5.0, 5.0 );
					connect( tmp, SIGNAL( valueChanged() ), this, SLOT( verticalChanged() ) );
				}

				_config->addParameter( "horizontal/samplinginterval", 1/4.0, "Sampling interval in units of nanoseconds" );
				if ( _devicename == "DC440" )
					_config->getParameter( "horizontal/samplinginterval" )->setRange( 1000.0/420, 1000 );
				_config->addParameter( "horizontal/samples", 100 );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"MaxSamplesPerChannel", &tmp ) );
				_config->getParameter( "horizontal/samples" )->setRange( 16, tmp );
				_config->addParameter( "horizontal/segments", 1, "Number of segments to acquire in one run" );
				_config->getParameter( "horizontal/segments" )->setRange( 1, int( pow( 2, 12 ) ) );
				_config->addParameter( "horizontal/delaytime", 0.0, "The time the trigger is delayed (in nanoseconds)." );
				_config->getParameter( "horizontal/delaytime" )->setRange( -1000.0, 1000.0 );
				connect( _config->getParameter( "horizontal" ), SIGNAL( valueChanged() ), this, SLOT( horizontalChanged() ) );

				_config->addParameter( "trigger/channel", 1, "Positive numbers for the input-channels, negative numbers for the external triggers." ); // TODO enum
				_config->getParameter( "trigger/channel" )->setRange( int( -_externaltriggers ), int( _channels ) );
				_config->addParameter( "trigger/slope", QVariant( 0 ), "0: Positive <br>1: Negative <br>2: Out of Window <br>3: Into Window <br>4: HF Divide" );
				_config->getParameter( "trigger/slope" )->setRange( 0, 4 );
				_config->addParameter( "trigger/level", 0.0 );
				_config->getParameter( "trigger/level" )->setGuiHint( NSCore::NamedParameterTreeItem::ReadAllowed );
				_config->addParameter( "trigger/levelintern", 0.0, "The trigger-level for internal triggermode in percent." );
				_config->getParameter( "trigger/levelintern" )->setRange( -100.0, 100.0 );
				connect( _config->getParameter( "trigger/levelintern" ), SIGNAL( valueChanged( double ) ), this, SLOT( triggerLevelIntern( double ) ) );
				_config->addParameter( "trigger/levelextern", 0.0, "The trigger-level for external triggers in Volts." );
				_config->getParameter( "trigger/levelextern" )->setRange( -5.0, 5.0 );
				connect( _config->getParameter( "trigger/levelextern" ), SIGNAL( valueChanged( double ) ), this, SLOT( triggerLevelExtern( double ) ) );
				_config->addParameter( "trigger/level2", 0.0 );
				_config->getParameter( "trigger/level2" )->setRange( -100.0, 100.0 );
				_config->addParameter( "trigger/coupling", QVariant::Int,
					"0: DC <br>1: AC <br>2: HF Reject <br>3: DC 50Ohm <br>4: AC 50Ohm" );
				_config->getParameter( "trigger/coupling" )->setRange( 0, 4 );
				connect( _config->getParameter( "trigger" ), SIGNAL( valueChanged() ), this, SLOT( triggerChanged() ) );

				//_config->addParameter( "external io", QVariant::Invalid );
				//_config->addParameter( "external io/IO A", QVariant::Int, "0: Disable <br>Input:<br>6: Enable Trigger <br>8: Prepare for trigger <br>Output:<br>19: 10MHz Reference Clock <br>20: Acquisition skips to next segment <br>21: Acquisition is active <br>22: Trigger is armed" );
				//_config->addParameter( "external io/IO B", QVariant::Int, "0: Disable <br>Input:<br>6: Enable Trigger <br>8: Prepare for trigger <br>Output:<br>19: 10MHz Reference Clock <br>20: Acquisition skips to next segment <br>21: Acquisition is active <br>22: Trigger is armed" );
				_config->addParameter( "filter", QVariant::Bool );
				_config->addParameter( "filter/minimum/min", -INT_MAX );
				_config->addParameter( "filter/minimum/max",  INT_MAX );
				_config->addParameter( "filter/maximum/min", -INT_MAX );
				_config->addParameter( "filter/maximum/max",  INT_MAX );
				connect( _config->getParameter( "filter" ), SIGNAL( valueChanged() ), this, SLOT( filterChanged() ) );

				_config->addParameter( "status/temperatur", 0.0 );
				_temperatur_param = _config->getParameter( "status/temperatur" );
				_temperatur_param->setGuiHint( NSCore::NamedParameterTreeItem::ReadAllowed );
				_temperatur_time.start();
				_config->addParameter( "status/counts per second", 0.0 );
				_statistics_cps_param = _config->getParameter( "status/counts per second" );
				_statistics_cps_param->setGuiHint( NSCore::NamedParameterTreeItem::ReadAllowed );
				_statistics_time.start();

				_last_cleanup.start();

			}
			~DigitizerThread() {
				qDebug() << "DigitizerThread::~DigitizerThread()";
				DigitizerDataPool::the()->houseKeeping( 0, 0 );
			}

			void stopThread() { _quit = true; }

			void startAcqisition() {
				qDebug() << "DigitizerThread::startAcqisition() setting _run=true";
				_run = true;
			}
			void stopAcqisition() {
				qDebug() << "DigitizerThread::stopAcqisition() setting _run=false";
				_run = false;
			}
			void setSingleShot( bool n ) { _single = n; }
			void scheduleCalibration() { _scheduleCalibration = true; }

			int channels() const { return _channels; }
			int externalTriggers() const { return _externaltriggers; }

			/**
			 * @brief Get and release data from the queue
			 */
			DigitizerData* first() {
				DigitizerData* tmp = 0;
				if ( _datamutex->tryLock( 100 ) ) {
					if ( !_data.empty() )
						tmp = _data.dequeue();
					_datamutex->unlock();
				}
				return tmp;
			 }

		private slots:
			/**
			 * @{
			 *
			 * @brief Settings have changed.
			 */
			void verticalChanged() { _verticalChanged = true; }
			void horizontalChanged() { _horizontalChanged = true; }
			void triggerChanged() { _triggerChanged = true; }
			void filterChanged() { _filterChanged = true; }
			// @}

			/// Set the trigger level from the internal trigger level.
			void triggerLevelIntern( double n ) {
				if ( _config->getParameter( "trigger/channel" )->toInt() > 0 )
					_config->getParameter( "trigger/level" )->setValue( n );
			}
			/// Set the trigger level from the external trigger level.
			void triggerLevelExtern( double n ) {
				if ( _config->getParameter( "trigger/channel" )->toInt() < 0 )
					_config->getParameter( "trigger/level" )->setValue( n*1000 );
			}

		private:
			/**
			 * @brief The real working function.
			 *
			 * Is run in the new thread...
			 */
			void run() {
				qDebug() << endl << "DigitizerThread is now running!";

				errorMessage( AcqrsD1_calibrate( _inst ), "AcqrsD1_calibrate" );

				errorMessage( AcqrsD1_configControlIO( _inst, 1, 19, 0, 0.0 ), "AcqrsD1_configControlIO" );

				int filter_min_min=-INT_MAX, filter_min_max=INT_MAX;
				int filter_max_min=-INT_MAX, filter_max_max=INT_MAX;
				bool filter=false;

				while ( !_quit ) {

					/*
					 * House-keeping of the data-pool.
					 */
					if ( _last_cleanup.elapsed() > 10000 ) {
						DigitizerDataPool::the()->houseKeeping( _last_used_size );
						_last_cleanup.restart();
					}

					/*
					 * Read the temperature of the first physical device every 5 seconds.
					 */
					if ( _temperatur_time.elapsed() > 5000 ) {
						qint32 temp;
						errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"Temperature 0", &temp ) );
						_temperatur_param->setValue( temp*1.0 );
						_temperatur_time.restart();
					}
					/*
					 * Do statistics every second.
					 */
					if ( _statistics_time.elapsed() > 1000 ) {
						_statistics_cps_param->setValue( _statistics_cps );
						_statistics_cps = 0.0;
						_statistics_time.restart();
					}

					if ( _run == false ) {
						/*
						 * If there is no acquisition running, we can just sleep...
						 */
						usleep( 1000 );
					} else {
						/*
						 * There are a lot of things to do when an acquisition should be done.
						 */

						//qDebug() << " * DigitizerThread: checking requisits";

						// First check wether the vertical parameters changed and set them accordingly.
						if ( _verticalChanged ) {
							for ( int i=-_externaltriggers; i<=_channels; ++i )
								if ( i!= 0 )
									configureChannel( i );
							_verticalChanged = false;
						}
						// Then check wether the horizontal parameters changed and set them accordingly.
						if ( _horizontalChanged ) {
							errorMessage( AcqrsD1_configHorizontal( _inst,
								_config->getParameter( "horizontal/samplinginterval" )->toDouble() * 1E-9,
								_config->getParameter( "horizontal/delaytime" )->toDouble() * 1E-9 ) );
							errorMessage( AcqrsD1_configMemory( _inst,
								_config->getParameter( "horizontal/samples" )->toInt(),
								_config->getParameter( "horizontal/segments" )->toInt() ) );
							_horizontalChanged = false;
						}
						// The trigger settings can also change and need to be set.
						if ( _triggerChanged ) {
							errorMessage( AcqrsD1_configTrigClass( _inst, 0, triggerSourceToPattern( _config->getParameter( "trigger/channel" )->toInt() ), 0, 0, 0, 0 ) );
							errorMessage( AcqrsD1_configTrigSource( _inst,
								_config->getParameter( "trigger/channel" )->toInt(),
								_config->getParameter( "trigger/coupling" )->toInt(),
								_config->getParameter( "trigger/slope" )->toInt(),
								_config->getParameter( "trigger/level" )->toDouble(),
								_config->getParameter( "trigger/level2" )->toDouble() ) );
							_triggerChanged = false;
						}
						// Is a calibration wanted?
						if ( _scheduleCalibration ) {
							errorMessage( AcqrsD1_calibrate( _inst ), "AcqrsD1_calibrate" );
							_scheduleCalibration = false;
						}
						// Did a filter-setting change?
						if ( _filterChanged ) {
							filter = _config->getParameter( "filter" )->toBool();
							filter_min_min = _config->getParameter( "filter/minimum/min" )->toInt();
							filter_min_max = _config->getParameter( "filter/minimum/max" )->toInt();
							filter_max_min = _config->getParameter( "filter/maximum/min" )->toInt();
							filter_max_max = _config->getParameter( "filter/maximum/max" )->toInt();
							_filterChanged = false;
						}

						//qDebug() << " * DigitizerThread: starting the acquisition";

						/*
						 * Do the real acquisition.
						 */
						errorMessage( AcqrsD1_acquire( _inst ), "AcqrsD1_acquire" );
						ViStatus status = AcqrsD1_waitForEndOfAcquisition( _inst, _config->getParameter( "timeout" )->toInt() );
						errorMessage( status, "AcqrsD1_waitForEndOfAcquisition" );

						//qDebug() << " * DigitizerThread: waitForEndOfAcquisition returned" << status;

						// Timeout-handling
						if ( status == signed( ACQIRIS_ERROR_ACQ_TIMEOUT ) ) {
							status = AcqrsD1_stopAcquisition( _inst );
							errorMessage( status );
							/*if ( _config->getParameter( "forcetrigger" )->toBool() )
								errorMessage( AcqrsD1_waitForEndOfAcquisition( _inst, 500 ) );
							else
								_run = false;*/
						}

						ViBoolean tmp;
						errorMessage( AcqrsD1_acqDone( _inst, &tmp ) );
						//qDebug() << " * DigitizerThread: acqDone returns" << bool( tmp );

						/*
						 * Get the data from the digitizer and store it in the outgoing buffer.
						 */
						long currentpad;
						errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"TbSegmentPad", &currentpad ), "AcqrsD1_getInstrumentInfo" );

						AqReadParameters readPar;
						readPar.flags = 0;
						readPar.dataType = _datatype_to_read;
						readPar.readMode = 1;
						readPar.firstSegment = 0;
						readPar.firstSampleInSeg = 0;
						errorMessage( AcqrsD1_getMemory( _inst, &( readPar.nbrSamplesInSeg ), &( readPar.nbrSegments ) ), "AcqrsD1_getMemory" );
						readPar.segmentOffset = readPar.nbrSamplesInSeg;
						readPar.segDescArraySize = sizeof( AqSegmentDescriptor ) * readPar.nbrSegments;
						readPar.dataArraySize = ( readPar.nbrSamplesInSeg * int( pow( 2, readPar.dataType ) ) + currentpad ) * ( 1+readPar.nbrSegments );
						readPar.dataArraySize = qMax( readPar.dataArraySize, ( readPar.nbrSamplesInSeg + currentpad ) * ( readPar.nbrSegments + 1 ) * int( pow( 2, readPar.dataType ) ) );
						//qDebug() << " Will allocate buffer of size" << readPar.dataArraySize << "bytes (=" << readPar.nbrSamplesInSeg << "*" << int( pow( 2, readPar.dataType ) ) << "+" << currentpad << "*" << ( 1+readPar.nbrSegments ) << ")";

						DigitizerData* data = DigitizerDataPool::the()->getData( _channels, readPar.nbrSegments, readPar.dataArraySize, pow( 2, readPar.dataType ) );

						for ( int i=0; i<_channels; ++i )
							errorMessage( AcqrsD1_readData( _inst, i+1, &readPar, data->data[ i ].dataArray, data->data[ i ].dataDesc, data->data[ i ].segDescArray ), "AcqrsD1_readData" );

						// Filter the data if wanted.
						if ( filter ) {
							data->filter = filter;
							data->filter_min_min = filter_min_min;
							data->filter_min_max = filter_min_max;
							data->filter_max_min = filter_max_min;
							data->filter_max_max = filter_max_max;
						}

						if ( !_digitizer->consumersIsEmpty() ) {
							_datamutex->lock();
							_data.enqueue( data );
							_datamutex->unlock();
						} else
							DigitizerDataPool::the()->deleteData( data );

						_statistics_cps += readPar.nbrSegments;

						/*
						 * Check wether a next acquisition is wanted or not.
						 */
						_acquisitionscounter->setValue( QVariant( int( _acquisitionscounter->toInt() + readPar.nbrSegments ) ) );
						if ( _acquisitionscounter_max->toInt() > 0 && _acquisitionscounter->toInt() > _acquisitionscounter_max->toInt() )
							_run = false;

						if ( _single || _quit )
							_run = false;
					}
				}
				qDebug() << "DigitizerThread finished!" << endl;
			}

			/**
			 * @brief Configuration of channels.
			 *
			 * Fits for both normal channels and external trigger inputs.
			 */
			void configureChannel( int ch ) {
				NSCore::NamedParameterTreeItem* tmp = _config->getParameter( QString( "Channel %1" ).arg( ch ) );
				errorMessage( AcqrsD1_configVertical( _inst, ch,
					fullscale( tmp->getParameter( "fullscale" )->toString() ),
					tmp->getParameter( "offset" )->toDouble(),
					( tmp->getParameter( "coupling" )?tmp->getParameter( "coupling" )->toInt():3 ),
					( tmp->getParameter( "bandwidth" )?tmp->getParameter( "bandwidth" )->toInt():0 )
					), QString( "configureChannel( %1 )" ).arg( ch ) );
#if 0
				// TODO Read back the values and change the gui accordingly?
				ViReal64 fs = 0, offset = 0;
				ViInt32 coupling = 3, bandwidth = 0;
				errorMessage( AcqrsD1_getVertical( _inst, ch, &fs, &offset, &coupling, &bandwidth ) );
				/*if ( fs != fullscale( tmp->getParameter( "fullscale" )->toString() ) )
					qDebug() << " Fullscale adopted";
				if ( offset != tmp->getParameter( "offset" )->toDouble() )
					qDebug() << " Offset adopted";*/
				qDebug() << " Fullscale is now" << fs;
				qDebug() << " Offset is now" << offset;
				qDebug() << " Coupling is now" << coupling;
				qDebug() << " Bandwidth is now" << bandwidth;
#endif
			}

			/**
			 * @brief Convert a fullscale-string to its double value.
			 */
			double fullscale( QString n ) const {
				float v=5.0;
				bool ok=false;
				if ( n.contains( "mV" ) )
					v = n.left( n.indexOf( "mV" ) ).toFloat( &ok )/1000;
				else if ( n.contains( "V" ) )
					v = n.left( n.indexOf( "V" ) ).toFloat( &ok );
				if ( ok )
					return v;
				return 5.0;
			}
			/**
			 * @brief Convert a trigger-channel to the aquivalent pattern.
			 */
			long triggerSourceToPattern( int channel ) const {
				int nbrExtTrigsPerModule=0, nbrIntTrigsPerModule=0, nbrModules=0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrModulesInInstrument", &nbrModules ) );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrInternalTriggers", &nbrIntTrigsPerModule ) );
				nbrIntTrigsPerModule = nbrIntTrigsPerModule / nbrModules;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrExternalTriggers", &nbrExtTrigsPerModule ) );
				nbrExtTrigsPerModule = nbrExtTrigsPerModule / nbrModules;
				int srcpattern=0;
				long moduleNbr, inputNbr;
				if ( channel > 0 && nbrIntTrigsPerModule > 0 ) {
					//qDebug( "\t Internal" );
					moduleNbr = ( channel - 1 ) / nbrIntTrigsPerModule;
					inputNbr = ( channel - 1 ) % nbrIntTrigsPerModule;
					srcpattern = ( moduleNbr<<16 ) + ( 0x1<<inputNbr );
				} else if ( channel < 0 && nbrExtTrigsPerModule > 0 ) {
					//qDebug( "\t External" );
					channel = -channel;
					moduleNbr = ( channel - 1 ) / nbrExtTrigsPerModule;
					inputNbr = ( channel - 1 ) % nbrExtTrigsPerModule;
					srcpattern = ( moduleNbr<<16 ) + ( 0x80000000>>inputNbr );
				}
				return srcpattern;
			}

			void errorMessage( ViStatus status, QString comment="" ) const { NSAcqiris::errorMessage( _inst, status, comment ); }
	};


	class ConverterThread : public QThread
	{
		Q_OBJECT

		DigitizerThread* _digitizer;

		bool _stop;

		int min( int t1, char t2 )   { return ( t2<t1 )?t2:t1; }
		int min( int t1, qint16 t2 ) { return ( t2<t1 )?t2:t1; }
		int max( int t1, char t2 )   { return ( t2>t1 )?t2:t1; }
		int max( int t1, qint16 t2 ) { return ( t2>t1 )?t2:t1; }
		public:
			ConverterThread( DigitizerThread* digitizer, QObject* p )
				: QThread( p )
				, _digitizer( digitizer )
				, _stop( false )
			{
			}

			void run() {
				while ( ! _stop ) {
					DigitizerData* tmp = _digitizer->first();
					if ( tmp == 0 )
						msleep( 500 );
					else {
						for ( int s=0; s<tmp->segments; ++s ) {
							bool pass=true;
							if ( tmp->filter ) {
								for ( int i=0; i<tmp->channels && pass; ++i ) {
									int rawmax=-INT_MAX, rawmin=INT_MAX;
									DigitizerChannelData* ch = &( tmp->data[ i ] );

									int offset = s * ch->dataDesc->returnedSamplesPerSeg;
									if ( tmp->dataBytes == 1 ) {
										char* d = static_cast<char*>( ch->dataArray );
										for ( int j=0; j<ch->dataDesc->returnedSamplesPerSeg; ++j ) {
											rawmax = max( rawmax, d[ offset + j ] );
											rawmin = min( rawmin, d[ offset + j ] );
										}
									} else {
										qint16* d = static_cast<qint16*>( ch->dataArray );
										for ( int j=0; j<ch->dataDesc->returnedSamplesPerSeg; ++j ) {
											rawmax = max( rawmax, qFromLittleEndian<qint16>( d[ offset + j ] ) );
											rawmin = min( rawmin, qFromLittleEndian<qint16>( d[ offset + j ] ) );
										}
									}
									pass = false;
									if ( rawmin > tmp->filter_min_min && rawmin < tmp->filter_min_max && rawmax > tmp->filter_max_min && rawmax < tmp->filter_max_max )
										pass = true;
								}
							}

							if ( pass ) {
								NSCore::AcqisitionDataSet data( tmp->channels, tmp->data[ 0 ].dataDesc->returnedSamplesPerSeg );

								data.addResult(  "horizontalposition", static_cast<AqSegmentDescriptor*>( tmp->data[ 0 ].segDescArray )[ s ].horPos );
								for ( int i=0; i<tmp->channels; ++i ) {
									DigitizerChannelData* ch = &( tmp->data[ i ] );
									double vGain = ch->dataDesc->vGain;
									double vOffset = ch->dataDesc->vOffset;
									int rawmax=-INT_MAX, rawmin=INT_MAX;

									double* realData = data.getRawChannel( i );
									int offset = s * ch->dataDesc->returnedSamplesPerSeg;
									if ( tmp->dataBytes == 1 ) {
										char* d = static_cast<char*>( ch->dataArray );
										for ( int j=0; j<data.samples(); ++j ) {
											rawmax = max( rawmax, d[ offset + j ] );
											rawmin = min( rawmin, d[ offset + j ] );
											realData[ j ] = d[ offset + j ] * vGain - vOffset;
										}
									} else {
										qint16* d = static_cast<qint16*>( ch->dataArray );
										for ( int j=0; j<data.samples(); ++j ) {
											rawmax = max( rawmax, qFromLittleEndian<qint16>( d[ offset + j ] ) );
											rawmin = min( rawmin, qFromLittleEndian<qint16>( d[ offset + j ] ) );
											realData[ j ] = ( qFromLittleEndian<qint16>( d[ offset + j ] ) ) * vGain - vOffset;
										}
									}

									data.addResult( QString( "channel %1/offset" ).arg( i ), vOffset );
									data.addResult( QString( "channel %1/fullscale" ).arg( i ), vGain * pow( 2, 8*tmp->dataBytes ) );
									data.addResult( QString( "channel %1/rawmin" ).arg( i ), rawmin );
									data.addResult( QString( "channel %1/rawmax" ).arg( i ), rawmax );
								}
								data.updateMinMax();
								emit newData( data );
							}
						}
						DigitizerDataPool::the()->deleteData( tmp );
					}
				}
			}

			void stopThread() { _stop = true; }
		signals:
			void newData( NSCore::AcqisitionDataSet );
	};

}; // namespace NSAcqiris

#endif // ACQIRIS_THREAD_H

